import pytest

from pythermalcomfort.models import vertical_tmp_grad_ppd


def test_vertical_tmp_grad_ppd():
    assert (
        vertical_tmp_grad_ppd(77, 77, 0.328, 50, 1.2, 0.5, 7 / 1.8, units="ip")[
            "PPD_vg"
        ]
    ) == 13.0
    assert (
        vertical_tmp_grad_ppd(77, 77, 0.328, 50, 1.2, 0.5, 7 / 1.8, units="ip")[
            "Acceptability"
        ]
    ) == False
    assert (vertical_tmp_grad_ppd(25, 25, 0.1, 50, 1.2, 0.5, 7)["PPD_vg"]) == 12.6
    assert (vertical_tmp_grad_ppd(25, 25, 0.1, 50, 1.2, 0.5, 4)["PPD_vg"]) == 1.7
    assert (
        vertical_tmp_grad_ppd(25, 25, 0.1, 50, 1.2, 0.5, 4)["Acceptability"]
    ) == True

    with pytest.raises(ValueError):
        vertical_tmp_grad_ppd(25, 25, 0.3, 50, 1.2, 0.5, 7)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import os
import sys

sys.path.insert(0, os.path.abspath(".."))

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.extlinks",
    "sphinx.ext.ifconfig",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx_rtd_theme",
]
source_suffix = ".rst"
master_doc = "index"
project = "pythermalcomfort"
year = "2019"
author = "Federico Tartarini"
copyright = "{0}, {1}".format(year, author)
version = release = "2.10.0"

autodoc_typehints = "none"
pygments_style = "trac"
templates_path = ["."]
extlinks = {
    "issue": (
        "https://github.com/CenterForTheBuiltEnvironment/pythermalcomfort/issues/%s",
        "issue %s",
    ),
    "pr": (
        "https://github.com/CenterForTheBuiltEnvironment/pythermalcomfort/pull/%s",
        "PR %s",
    ),
}
# on_rtd is whether we are on readthedocs.org
on_rtd = os.environ.get("READTHEDOCS", None) == "True"

html_theme = "sphinx_rtd_theme"

html_use_smartypants = True
html_last_updated_fmt = "%b %d, %Y"
html_split_index = False
html_sidebars = {
    "**": ["searchbox.html", "globaltoc.html", "sourcelink.html"],
}
html_short_title = "%s-%s" % (project, version)

napoleon_use_ivar = True
napoleon_use_rtype = False
napoleon_use_param = False



import numpy as np
import pytest

from pythermalcomfort.models import (
    utci,
)
from pythermalcomfort.models.utci import _utci_optimized


@pytest.fixture
def data_test_utci():
    return [  # I have commented the lines of code that don't pass the test
        {"tdb": 25, "tr": 27, "rh": 50, "v": 1, "return": {"utci": 25.2}},
        {"tdb": 19, "tr": 24, "rh": 50, "v": 1, "return": {"utci": 20.0}},
        {"tdb": 19, "tr": 14, "rh": 50, "v": 1, "return": {"utci": 16.8}},
        {"tdb": 27, "tr": 22, "rh": 50, "v": 1, "return": {"utci": 25.5}},
        {"tdb": 27, "tr": 22, "rh": 50, "v": 10, "return": {"utci": 20.0}},
        {"tdb": 27, "tr": 22, "rh": 50, "v": 16, "return": {"utci": 15.8}},
        {"tdb": 51, "tr": 22, "rh": 50, "v": 16, "return": {"utci": np.nan}},
        {"tdb": 27, "tr": 22, "rh": 50, "v": 0, "return": {"utci": np.nan}},
    ]


def test_utci(data_test_utci):
    for row in data_test_utci:
        np.testing.assert_equal(
            utci(row["tdb"], row["tr"], row["v"], row["rh"]),
            row["return"][list(row["return"].keys())[0]],
        )

    assert (utci(tdb=77, tr=77, v=3.28, rh=50, units="ip")) == 76.4

    assert (
        utci(tdb=30, tr=27, v=1, rh=50, units="si", return_stress_category=True)
    ) == {"utci": 29.6, "stress_category": "moderate heat stress"}
    assert (utci(tdb=9, tr=9, v=1, rh=50, units="si", return_stress_category=True)) == {
        "utci": 8.7,
        "stress_category": "slight cold stress",
    }


def test_utci_numpy(data_test_utci):
    tdb = np.array([d["tdb"] for d in data_test_utci])
    tr = np.array([d["tr"] for d in data_test_utci])
    rh = np.array([d["rh"] for d in data_test_utci])
    v = np.array([d["v"] for d in data_test_utci])
    expect = np.array([d["return"]["utci"] for d in data_test_utci])

    np.testing.assert_equal(utci(tdb, tr, v, rh), expect)

    tdb = np.array([25, 25])
    tr = np.array([27, 25])
    v = np.array([1, 1])
    rh = np.array([50, 50])
    expect = {
        "utci": np.array([25.2, 24.6]),
        "stress_category": np.array(["no thermal stress", "no thermal stress"]),
    }

    result = utci(tdb, tr, v, rh, units="si", return_stress_category=True)
    np.testing.assert_equal(result["utci"], expect["utci"])
    np.testing.assert_equal(result["stress_category"], expect["stress_category"])


def test_utci_optimized():
    np.testing.assert_equal(
        np.around(_utci_optimized([25, 27], 1, 1, 1.5), 2), [24.73, 26.57]
    )



from typing import Union, List

import numpy as np

from pythermalcomfort.psychrometrics import t_o
from pythermalcomfort.shared_functions import valid_range
from pythermalcomfort.utilities import (
    units_converter,
)


def adaptive_en(
    tdb: Union[float, int, np.ndarray, List[float], List[int]],
    tr: Union[float, int, np.ndarray, List[float], List[int]],
    t_running_mean: Union[float, int, np.ndarray, List[float], List[int]],
    v: Union[float, int, np.ndarray, List[float], List[int]],
    units: str = "SI",
    limit_inputs: bool = True,
):
    """Determines the adaptive thermal comfort based on EN 16798-1 2019 [3]_

    Parameters
    ----------
    tdb : float, int, or array-like
        dry bulb air temperature, default in [°C] in [°F] if `units` = 'IP'
    tr : float, int, or array-like
        mean radiant temperature, default in [°C] in [°F] if `units` = 'IP'
    t_running_mean: float, int, or array-like
        running mean temperature, default in [°C] in [°C] in [°F] if `units` = 'IP'

        The running mean temperature can be calculated using the function
        :py:meth:`pythermalcomfort.utilities.running_mean_outdoor_temperature`.
    v : float, int, or array-like
        air speed, default in [m/s] in [fps] if `units` = 'IP'

        Note: Indoor operative temperature correction is applicable for buildings equipped
        with fans or personal systems providing building occupants with personal
        control over air speed at occupant level.
        For operative temperatures above 25°C the comfort zone upper limit can be
        increased by 1.2 °C (0.6 < v < 0.9 m/s), 1.8 °C (0.9 < v < 1.2 m/s), 2.2 °C ( v
        > 1.2 m/s)
    units : {'SI', 'IP'}
        select the SI (International System of Units) or the IP (Imperial Units) system.
    limit_inputs : boolean default True
        By default, if the inputs are outsude the standard applicability limits the
        function returns nan. If False returns pmv and ppd values even if input values are
        outside the applicability limits of the model.

    Returns
    -------
    tmp_cmf : float, int, or array-like
        Comfort temperature at that specific running mean temperature, default in [°C]
        or in [°F]
    acceptability_cat_i : bol or array-like
        If the indoor conditions comply with comfort category I
    acceptability_cat_ii : bol or array-like
        If the indoor conditions comply with comfort category II
    acceptability_cat_iii : bol or array-like
        If the indoor conditions comply with comfort category III
    tmp_cmf_cat_i_up : float, int, or array-like
        Upper acceptable comfort temperature for category I, default in [°C] or in [°F]
    tmp_cmf_cat_ii_up : float, int, or array-like
        Upper acceptable comfort temperature for category II, default in [°C] or in [°F]
    tmp_cmf_cat_iii_up : float, int, or array-like
        Upper acceptable comfort temperature for category III, default in [°C] or in [°F]
    tmp_cmf_cat_i_low : float, int, or array-like
        Lower acceptable comfort temperature for category I, default in [°C] or in [°F]
    tmp_cmf_cat_ii_low : float, int, or array-like
        Lower acceptable comfort temperature for category II, default in [°C] or in [°F]
    tmp_cmf_cat_iii_low : float or array-like
        Lower acceptable comfort temperature for category III, default in [°C] or in [°F]

    Notes
    -----
    You can use this function to calculate if your conditions are within the EN
    adaptive thermal comfort region.
    Calculations with comply with the EN 16798-1 2019 [3]_.

    Examples
    --------
    .. code-block:: python

        >>> from pythermalcomfort.models import adaptive_en
        >>> results = adaptive_en(tdb=25, tr=25, t_running_mean=20, v=0.1)
        >>> print(results)
        {'tmp_cmf': 25.4, 'acceptability_cat_i': True, 'acceptability_cat_ii': True, ... }

        >>> print(results['acceptability_cat_i'])
        True
        # The conditions you entered are considered to comply with Category I

        >>> # for users who wants to use the IP system
        >>> results = adaptive_en(tdb=77, tr=77, t_running_mean=68, v=0.3, units='ip')
        >>> print(results)
        {'tmp_cmf': 77.7, 'acceptability_cat_i': True, 'acceptability_cat_ii': True, ... }

        >>> results = adaptive_en(tdb=25, tr=25, t_running_mean=9, v=0.1)
        {'tmp_cmf': nan, 'acceptability_cat_i': True, 'acceptability_cat_ii': True, ... }
        # The adaptive thermal comfort model can only be used
        # if the running mean temperature is between 10 °C and 30 °C
    """

    tdb = np.array(tdb)
    tr = np.array(tr)
    t_running_mean = np.array(t_running_mean)
    v = np.array(v)
    standard = "iso"

    if units.lower() == "ip":
        tdb, tr, t_running_mean, vr = units_converter(
            tdb=tdb, tr=tr, tmp_running_mean=t_running_mean, v=v
        )

    trm_valid = valid_range(t_running_mean, (10.0, 33.5))

    to = t_o(tdb, tr, v, standard=standard)

    # calculate cooling effect (ce) of elevated air speed when top > 25 degC.
    ce = np.where((v >= 0.6) & (to >= 25.0), 999, 0)
    ce = np.where((v < 0.9) & (ce == 999), 1.2, ce)
    ce = np.where((v < 1.2) & (ce == 999), 1.8, ce)
    ce = np.where(ce == 999, 2.2, ce)

    t_cmf = 0.33 * t_running_mean + 18.8

    if limit_inputs:
        all_valid = ~(np.isnan(trm_valid))
        t_cmf = np.where(all_valid, t_cmf, np.nan)

    t_cmf_i_lower = t_cmf - 3.0
    t_cmf_ii_lower = t_cmf - 4.0
    t_cmf_iii_lower = t_cmf - 5.0
    t_cmf_i_upper = t_cmf + 2.0 + ce
    t_cmf_ii_upper = t_cmf + 3.0 + ce
    t_cmf_iii_upper = t_cmf + 4.0 + ce

    acceptability_i = np.where(
        (t_cmf_i_lower <= to) & (to <= t_cmf_i_upper), True, False
    )
    acceptability_ii = np.where(
        (t_cmf_ii_lower <= to) & (to <= t_cmf_ii_upper), True, False
    )
    acceptability_iii = np.where(
        (t_cmf_iii_lower <= to) & (to <= t_cmf_iii_upper), True, False
    )

    if units.lower() == "ip":
        t_cmf, t_cmf_i_upper, t_cmf_ii_upper, t_cmf_iii_upper = units_converter(
            from_units="si",
            tmp_cmf=t_cmf,
            tmp_cmf_cat_i_up=t_cmf_i_upper,
            tmp_cmf_cat_ii_up=t_cmf_ii_upper,
            tmp_cmf_cat_iii_up=t_cmf_iii_upper,
        )
        t_cmf_i_lower, t_cmf_ii_lower, t_cmf_iii_lower = units_converter(
            from_units="si",
            tmp_cmf_cat_i_low=t_cmf_i_lower,
            tmp_cmf_cat_ii_low=t_cmf_ii_lower,
            tmp_cmf_cat_iii_low=t_cmf_iii_lower,
        )

    results = {
        "tmp_cmf": np.around(t_cmf, 1),
        "acceptability_cat_i": acceptability_i,
        "acceptability_cat_ii": acceptability_ii,
        "acceptability_cat_iii": acceptability_iii,
        "tmp_cmf_cat_i_up": np.around(t_cmf_i_upper, 1),
        "tmp_cmf_cat_ii_up": np.around(t_cmf_ii_upper, 1),
        "tmp_cmf_cat_iii_up": np.around(t_cmf_iii_upper, 1),
        "tmp_cmf_cat_i_low": np.around(t_cmf_i_lower, 1),
        "tmp_cmf_cat_ii_low": np.around(t_cmf_ii_lower, 1),
        "tmp_cmf_cat_iii_low": np.around(t_cmf_iii_lower, 1),
    }

    return results



import numpy as np

from pythermalcomfort.models import adaptive_en


def test_adaptive_en():
    np.testing.assert_equal(
        adaptive_en(
            tdb=[25, 25, 23.5], tr=[25, 25, 23.5], t_running_mean=[9, 20, 28], v=0.1
        ),
        {
            "tmp_cmf": [np.nan, 25.4, 28.0],
            "acceptability_cat_i": [False, True, False],
            "acceptability_cat_ii": [False, True, False],
            "acceptability_cat_iii": [False, True, True],
            "tmp_cmf_cat_i_up": [np.nan, 27.4, 30.0],
            "tmp_cmf_cat_ii_up": [np.nan, 28.4, 31.0],
            "tmp_cmf_cat_iii_up": [np.nan, 29.4, 32.0],
            "tmp_cmf_cat_i_low": [np.nan, 22.4, 25.0],
            "tmp_cmf_cat_ii_low": [np.nan, 21.4, 24.0],
            "tmp_cmf_cat_iii_low": [np.nan, 20.4, 23.0],
        },
    )



from dataclasses import dataclass
from typing import Union, List

import numpy as np

from pythermalcomfort.psychrometrics import t_o
from pythermalcomfort.shared_functions import valid_range
from pythermalcomfort.utilities import (
    units_converter,
    check_standard_compliance_array,
)


@dataclass
class AdaptiveASHRAE:
    tmp_cmf: Union[float, int, np.ndarray, List[float], List[int]]
    tmp_cmf_80_low: Union[float, int, np.ndarray, List[float], List[int]]
    tmp_cmf_80_up: Union[float, int, np.ndarray, List[float], List[int]]
    tmp_cmf_90_low: Union[float, int, np.ndarray, List[float], List[int]]
    tmp_cmf_90_up: Union[float, int, np.ndarray, List[float], List[int]]
    acceptability_80: Union[float, int, np.ndarray, List[float], List[int]]
    acceptability_90: Union[float, int, np.ndarray, List[float], List[int]]

    def __getitem__(self, item):
        return getattr(self, item)


def adaptive_ashrae(
    tdb: Union[float, int, np.ndarray, List[float], List[int]],
    tr: Union[float, int, np.ndarray, List[float], List[int]],
    t_running_mean: Union[float, int, np.ndarray, List[float], List[int]],
    v: Union[float, int, np.ndarray, List[float], List[int]],
    units: str = "SI",
    limit_inputs: bool = True,
) -> AdaptiveASHRAE:
    """Determines the adaptive thermal comfort based on ASHRAE 55. The adaptive
    model relates indoor design temperatures or acceptable temperature ranges
    to outdoor meteorological or climatological parameters. The adaptive model
    can only be used in occupant-controlled naturally conditioned spaces that
    meet all the following criteria:

    * There is no mechianical cooling or heating system in operation
    * Occupants have a metabolic rate between 1.0 and 1.5 met
    * Occupants are free to adapt their clothing within a range as wide as 0.5 and 1.0 clo
    * The prevailing mean (runnin mean) outdoor temperature is between 10 and 33.5 °C

    Parameters
    ----------
    tdb : float, int, or array-like
        dry bulb air temperature, default in [°C] in [°F] if `units` = 'IP'
    tr : float, int, or array-like
        mean radiant temperature, default in [°C] in [°F] if `units` = 'IP'
    t_running_mean: float, int, or array-like
        running mean temperature, default in [°C] in [°C] in [°F] if `units` = 'IP'

        The running mean temperature can be calculated using the function
        :py:meth:`pythermalcomfort.utilities.running_mean_outdoor_temperature`.
    v : float, int, or array-like
        air speed, default in [m/s] in [fps] if `units` = 'IP'
    units : str, optional
        select the SI (International System of Units) or the IP (Imperial Units) system.
        Supported values are 'SI' and 'IP'. Defaults to 'SI'.
    limit_inputs : boolean default True
        By default, if the inputs are outsude the standard applicability limits the
        function returns nan. If False returns pmv and ppd values even if input values are
        outside the applicability limits of the model.

        The ASHRAE 55 2020 limits are 10 < tdb [°C] < 40, 10 < tr [°C] < 40,
        0 < vr [m/s] < 2, 10 < t running mean [°C] < 33.5

    Returns
    -------
    tmp_cmf : float, int, or array-like
        Comfort temperature a that specific running mean temperature, default in [°C]
        or in [°F]
    tmp_cmf_80_low : float, int, or array-like
        Lower acceptable comfort temperature for 80% occupants, default in [°C] or in [°F]
    tmp_cmf_80_up : float, int, or array-like
        Upper acceptable comfort temperature for 80% occupants, default in [°C] or in [°F]
    tmp_cmf_90_low : float, int, or array-like
        Lower acceptable comfort temperature for 90% occupants, default in [°C] or in [°F]
    tmp_cmf_90_up : float, int, or array-like
        Upper acceptable comfort temperature for 90% occupants, default in [°C] or in [°F]
    acceptability_80 : bol or array-like
        Acceptability for 80% occupants
    acceptability_90 : bol or array-like
        Acceptability for 90% occupants

    Notes
    -----
    You can use this function to calculate if your conditions are within the `adaptive
    thermal comfort region`.
    Calculations with comply with the ASHRAE 55 2020 Standard [1]_.

    Examples
    --------
    .. code-block:: python

        >>> from pythermalcomfort.models import adaptive_ashrae
        >>> results = adaptive_ashrae(tdb=25, tr=25, t_running_mean=20, v=0.1)
        >>> print(results)
        {'tmp_cmf': 24.0, 'tmp_cmf_80_low': 20.5, 'tmp_cmf_80_up': 27.5,
        'tmp_cmf_90_low': 21.5, 'tmp_cmf_90_up': 26.5, 'acceptability_80': array(True),
        'acceptability_90': array(True)}

        >>> print(results.acceptability_80)  # or use print(results["acceptability_80"])
        True
        # The conditions you entered are considered to be comfortable for by 80% of the
        occupants

        >>> # You can also pass arrays as input to the function
        >>> results = adaptive_ashrae(tdb=[25, 26], tr=25, t_running_mean=20, v=0.1)
        >>> print(results)
        {'tmp_cmf': 24.0, 'tmp_cmf_80_low': 20.5, 'tmp_cmf_80_up': 27.5,
        'tmp_cmf_90_low': 21.5, 'tmp_cmf_90_up': 26.5, 'acceptability_80': array(True),
        'acceptability_90': array(True)}

        >>> # for users who want to use the IP system
        >>> results = adaptive_ashrae(tdb=77, tr=77, t_running_mean=68, v=0.3, units='ip')
        >>> print(results)
        {'tmp_cmf': 75.2, 'tmp_cmf_80_low': 68.9, 'tmp_cmf_80_up': 81.5,
        'tmp_cmf_90_low': 70.7, 'tmp_cmf_90_up': 79.7, 'acceptability_80': array(True),
        'acceptability_90': array(True)}

        >>> adaptive_ashrae(tdb=25, tr=25, t_running_mean=9, v=0.1)
        {'tmp_cmf': nan, 'tmp_cmf_80_low': nan, ... }
        # The adaptive thermal comfort model can only be used
        # if the running mean temperature is higher than 10°C
    """

    tdb = np.array(tdb)
    tr = np.array(tr)
    t_running_mean = np.array(t_running_mean)
    v = np.array(v)
    standard = "ashrae"

    # Validate units string
    valid_units: List[str] = ["SI", "IP"]
    if units.upper() not in valid_units:
        raise ValueError(f"Invalid unit: {units}. Supported units are {valid_units}.")

    if units.lower() == "ip":
        tdb, tr, t_running_mean, v = units_converter(
            tdb=tdb, tr=tr, tmp_running_mean=t_running_mean, v=v
        )

    (
        tdb_valid,
        tr_valid,
        v_valid,
    ) = check_standard_compliance_array(standard, tdb=tdb, tr=tr, v=v)
    trm_valid = valid_range(t_running_mean, (10.0, 33.5))

    to = t_o(tdb, tr, v, standard=standard)

    # calculate cooling effect (ce) of elevated air speed when top > 25 degC.
    ce = np.where((v >= 0.6) & (to >= 25.0), 999, 0)
    ce = np.where((v < 0.9) & (ce == 999), 1.2, ce)
    ce = np.where((v < 1.2) & (ce == 999), 1.8, ce)
    ce = np.where(ce == 999, 2.2, ce)

    # Relation between comfort and outdoor temperature
    t_cmf = 0.31 * t_running_mean + 17.8

    if limit_inputs:
        all_valid = ~(
            np.isnan(tdb_valid)
            | np.isnan(tr_valid)
            | np.isnan(v_valid)
            | np.isnan(trm_valid)
        )
        t_cmf = np.where(all_valid, t_cmf, np.nan)

    t_cmf = np.around(t_cmf, 1)

    tmp_cmf_80_low = t_cmf - 3.5
    tmp_cmf_90_low = t_cmf - 2.5
    tmp_cmf_80_up = t_cmf + 3.5 + ce
    tmp_cmf_90_up = t_cmf + 2.5 + ce

    acceptability_80 = np.where(
        (tmp_cmf_80_low <= to) & (to <= tmp_cmf_80_up), True, False
    )
    acceptability_90 = np.where(
        (tmp_cmf_90_low <= to) & (to <= tmp_cmf_90_up), True, False
    )

    if units.lower() == "ip":
        (
            t_cmf,
            tmp_cmf_80_low,
            tmp_cmf_80_up,
            tmp_cmf_90_low,
            tmp_cmf_90_up,
        ) = units_converter(
            from_units="si",
            tmp_cmf=t_cmf,
            tmp_cmf_80_low=tmp_cmf_80_low,
            tmp_cmf_80_up=tmp_cmf_80_up,
            tmp_cmf_90_low=tmp_cmf_90_low,
            tmp_cmf_90_up=tmp_cmf_90_up,
        )

    return AdaptiveASHRAE(
        tmp_cmf=t_cmf,
        tmp_cmf_80_low=tmp_cmf_80_low,
        tmp_cmf_80_up=tmp_cmf_80_up,
        tmp_cmf_90_low=tmp_cmf_90_low,
        tmp_cmf_90_up=tmp_cmf_90_up,
        acceptability_80=acceptability_80,
        acceptability_90=acceptability_90,
    )



from typing import Union, List
import math

import numpy as np
from numba import jit, vectorize, float64

from pythermalcomfort.psychrometrics import p_sat_torr
from pythermalcomfort.utilities import (
    body_surface_area,
)


def two_nodes(
    tdb: Union[float, int, np.ndarray, List[float], List[int]],
    tr: Union[float, int, np.ndarray, List[float], List[int]],
    v: Union[float, int, np.ndarray, List[float], List[int]],
    rh: Union[float, int, np.ndarray, List[float], List[int]],
    met: Union[float, int, np.ndarray, List[float], List[int]],
    clo: Union[float, int, np.ndarray, List[float], List[int]],
    wme: Union[float, int, np.ndarray, List[float], List[int]] = 0,
    body_surface_area=1.8258,
    p_atmospheric=101325,
    body_position="standing",
    max_skin_blood_flow=90,
    **kwargs,
):
    """Two-node model of human temperature regulation Gagge et al. (1986). [10]_
    This model it can be used to calculate a variety of indices,
    including:

    * Gagge's version of Fanger's Predicted Mean Vote (PMV). This function uses the Fanger's PMV equations but it replaces the heat loss and gain terms with those calculated by the two node model developed by Gagge et al. (1986) [10]_.
    * PMV SET and the predicted thermal sensation based on SET [10]_. This function is similar in all aspects to the :py:meth:`pythermalcomfort.models.pmv_gagge` however, it uses the :py:meth:`pythermalcomfort.models.set` equation to calculate the dry heat loss by convection.
    * Thermal discomfort (DISC) as the relative thermoregulatory strain necessary to restore a state of comfort and thermal equilibrium by sweating [10]_. DISC is described numerically as: comfortable and pleasant (0), slightly uncomfortable but acceptable (1), uncomfortable and unpleasant (2), very uncomfortable (3), limited tolerance (4), and intolerable (S). The range of each category is ± 0.5 numerically. In the cold, the classical negative category descriptions used for Fanger's PMV apply [10]_.
    * Heat gains and losses via convection, radiation and conduction.
    * The Standard Effective Temperature (SET)
    * The New Effective Temperature (ET)
    * The Predicted  Thermal  Sensation  (TSENS)
    * The Predicted  Percent  Dissatisfied  Due  to  Draft  (PD)
    * Predicted  Percent  Satisfied  With  the  Level  of  Air  Movement"   (PS)

    Parameters
    ----------
    tdb : float, int, or array-like
        dry bulb air temperature, default in [°C] in [°F] if `units` = 'IP'
    tr : float or array-like
        mean radiant temperature, default in [°C] in [°F] if `units` = 'IP'
    v : float or array-like
        air speed, default in [m/s] in [fps] if `units` = 'IP'
    rh : float or array-like
        relative humidity, [%]
    met : float or array-like
        metabolic rate, [met]
    clo : float or array-like
        clothing insulation, [clo]
    wme : float or array-like
        external work, [met] default 0
    body_surface_area : float
        body surface area, default value 1.8258 [m2] in [ft2] if `units` = 'IP'

        The body surface area can be calculated using the function
        :py:meth:`pythermalcomfort.utilities.body_surface_area`.
    p_atmospheric : float
        atmospheric pressure, default value 101325 [Pa] in [atm] if `units` = 'IP'
    body_position: str default="standing" or array-like
        select either "sitting" or "standing"
    max_skin_blood_flow : float
        maximum blood flow from the core to the skin, [kg/h/m2] default 80

    Other Parameters
    ----------------
    round: boolean, default True
        if True rounds output values, if False it does not round them
    max_sweating : float
        Maximum rate at which regulatory sweat is generated, [kg/h/m2]
    w_max : float
        Maximum skin wettedness (w) adimensional. Ranges from 0 and 1.

    Returns
    -------
    e_skin : float or array-like
        Total rate of evaporative heat loss from skin, [W/m2]. Equal to e_rsw + e_diff
    e_rsw : float or array-like
        Rate of evaporative heat loss from sweat evaporation, [W/m2]
    e_diff : float or array-like
        Rate of evaporative heat loss from moisture diffused through the skin, [W/m2]
    e_max : float or array-like
        Maximum rate of evaporative heat loss from skin, [W/m2]
    q_sensible : float or array-like
        Sensible heat loss from skin, [W/m2]
    q_skin : float or array-like
        Total rate of heat loss from skin, [W/m2]. Equal to q_sensible + e_skin
    q_res : float or array-like
        Total rate of heat loss through respiration, [W/m2]
    t_core : float or array-like
        Core temperature, [°C]
    t_skin : float or array-like
        Skin temperature, [°C]
    m_bl : float or array-like
        Skin blood flow, [kg/h/m2]
    m_rsw : float or array-like
        Rate at which regulatory sweat is generated, [kg/h/m2]
    w : float or array-like
        Skin wettedness, adimensional. Ranges from 0 and 1.
    w_max : float or array-like
        Skin wettedness (w) practical upper limit, adimensional. Ranges from 0 and 1.
    set : float or array-like
        Standard Effective Temperature (SET)
    et : float or array-like
        New Effective Temperature (ET)
    pmv_gagge : float or array-like
        PMV Gagge
    pmv_set : float or array-like
        PMV SET
    pd : float or array-like
        Predicted  Percent  Dissatisfied  Due  to  Draft"
    ps : float or array-like
        Predicted  Percent  Satisfied  With  the  Level  of  Air  Movement
    disc : float or array-like
        Thermal discomfort
    t_sens : float or array-like
        Predicted  Thermal  Sensation

    Examples
    --------
    .. code-block:: python

        >>> from pythermalcomfort.models import two_nodes
        >>> print(two_nodes(tdb=25, tr=25, v=0.3, rh=50, met=1.2, clo=0.5))
        {'e_skin': 15.8, 'e_rsw': 6.5, 'e_diff': 9.3, ... }
        >>> print(two_nodes(tdb=[25, 25], tr=25, v=0.3, rh=50, met=1.2, clo=0.5))
        {'e_skin': array([15.8, 15.8]), 'e_rsw': array([6.5, 6.5]), ... }
    """
    default_kwargs = {
        "round": True,
        "calculate_ce": False,
        "max_sweating": 500,
        "w_max": False,
    }
    kwargs = {**default_kwargs, **kwargs}

    tdb = np.array(tdb)
    tr = np.array(tr)
    v = np.array(v)
    rh = np.array(rh)
    met = np.array(met)
    clo = np.array(clo)
    wme = np.array(wme)
    body_position = np.array(body_position)

    vapor_pressure = rh * p_sat_torr(tdb) / 100

    if kwargs["calculate_ce"]:
        result = _two_nodes_optimized_return_set(
            tdb,
            tr,
            v,
            met,
            clo,
            vapor_pressure,
            wme,
            body_surface_area,
            p_atmospheric,
            1,
        )
        return {"_set": result}

    (
        _set,
        e_skin,
        e_rsw,
        e_max,
        q_sensible,
        q_skin,
        q_res,
        t_core,
        t_skin,
        m_bl,
        m_rsw,
        w,
        w_max,
        et,
        pmv_gagge,
        pmv_set,
        disc,
        t_sens,
    ) = np.vectorize(_two_nodes_optimized, cache=True)(
        tdb=tdb,
        tr=tr,
        v=v,
        met=met,
        clo=clo,
        vapor_pressure=vapor_pressure,
        wme=wme,
        body_surface_area=body_surface_area,
        p_atmospheric=p_atmospheric,
        body_position=body_position,
        calculate_ce=kwargs["calculate_ce"],
        max_skin_blood_flow=max_skin_blood_flow,
        max_sweating=kwargs["max_sweating"],
        w_max=kwargs["w_max"],
    )

    output = {
        "e_skin": e_skin,
        "e_rsw": e_rsw,
        "e_max": e_max,
        "q_sensible": q_sensible,
        "q_skin": q_skin,
        "q_res": q_res,
        "t_core": t_core,
        "t_skin": t_skin,
        "m_bl": m_bl,
        "m_rsw": m_rsw,
        "w": w,
        "w_max": w_max,
        "_set": _set,
        "et": et,
        "pmv_gagge": pmv_gagge,
        "pmv_set": pmv_set,
        "disc": disc,
        "t_sens": t_sens,
    }

    for key in output.keys():
        # round the results if needed
        if kwargs["round"]:
            output[key] = np.around(output[key], 1)

    return output


@jit(nopython=True)
def _two_nodes_optimized(
    tdb,
    tr,
    v,
    met,
    clo,
    vapor_pressure,
    wme,
    body_surface_area,
    p_atmospheric,
    body_position,
    calculate_ce=False,
    max_skin_blood_flow=90,
    max_sweating=500,
    w_max=False,
):
    # Initial variables as defined in the ASHRAE 55-2020
    air_speed = max(v, 0.1)
    k_clo = 0.25
    body_weight = 70  # body weight in kg
    met_factor = 58.2  # met conversion factor
    sbc = 0.000000056697  # Stefan-Boltzmann constant (W/m2K4)
    c_sw = 170  # driving coefficient for regulatory sweating
    c_dil = 120  # driving coefficient for vasodilation ashrae says 50 see page 9.19
    c_str = 0.5  # driving coefficient for vasoconstriction

    temp_skin_neutral = 33.7
    temp_core_neutral = 36.8
    alfa = 0.1
    temp_body_neutral = alfa * temp_skin_neutral + (1 - alfa) * temp_core_neutral
    skin_blood_flow_neutral = 6.3

    t_skin = temp_skin_neutral
    t_core = temp_core_neutral
    m_bl = skin_blood_flow_neutral

    # initialize some variables
    e_skin = 0.1 * met  # total evaporative heat loss, W
    q_sensible = 0  # total sensible heat loss, W
    w = 0  # skin wettedness
    _set = 0  # standard effective temperature
    e_rsw = 0  # heat lost by vaporization sweat
    e_diff = 0  # vapor diffusion through skin
    e_max = 0  # maximum evaporative capacity
    m_rsw = 0  # regulatory sweating
    q_res = 0  # heat loss due to respiration
    et = 0  # effective temperature
    e_req = 0  # evaporative heat loss required for tmp regulation
    r_ea = 0
    r_ecl = 0
    c_res = 0  # convective heat loss respiration

    pressure_in_atmospheres = p_atmospheric / 101325
    length_time_simulation = 60  # length time simulation
    n_simulation = 0

    r_clo = 0.155 * clo  # thermal resistance of clothing, C M^2 /W
    f_a_cl = 1.0 + 0.15 * clo  # increase in body surface area due to clothing
    lr = 2.2 / pressure_in_atmospheres  # Lewis ratio
    rm = (met - wme) * met_factor  # metabolic rate
    m = met * met_factor  # metabolic rate

    e_comfort = 0.42 * (rm - met_factor)  # evaporative heat loss during comfort
    if e_comfort < 0:
        e_comfort = 0

    i_cl = 1.0  # permeation efficiency of water vapour naked skin
    if clo > 0:
        i_cl = 0.45  # permeation efficiency of water vapour through the clothing layer

    if not w_max:  # if the user did not pass a value of w_max
        w_max = 0.38 * pow(air_speed, -0.29)  # critical skin wettedness naked
        if clo > 0:
            w_max = 0.59 * pow(air_speed, -0.08)  # critical skin wettedness clothed

    # h_cc corrected convective heat transfer coefficient
    h_cc = 3.0 * pow(pressure_in_atmospheres, 0.53)
    # h_fc forced convective heat transfer coefficient, W/(m2 °C)
    h_fc = 8.600001 * pow((air_speed * pressure_in_atmospheres), 0.53)
    h_cc = max(h_cc, h_fc)
    if not calculate_ce and met > 0.85:
        h_c_met = 5.66 * (met - 0.85) ** 0.39
        h_cc = max(h_cc, h_c_met)

    h_r = 4.7  # linearized radiative heat transfer coefficient
    h_t = h_r + h_cc  # sum of convective and radiant heat transfer coefficient W/(m2*K)
    r_a = 1.0 / (f_a_cl * h_t)  # resistance of air layer to dry heat
    t_op = (h_r * tr + h_cc * tdb) / h_t  # operative temperature

    t_body = alfa * t_skin + (1 - alfa) * t_core  # mean body temperature, °C

    # respiration
    q_res = 0.0023 * m * (44.0 - vapor_pressure)  # latent heat loss due to respiration
    c_res = 0.0014 * m * (34.0 - tdb)  # sensible convective heat loss respiration

    while n_simulation < length_time_simulation:

        n_simulation += 1

        iteration_limit = 150  # for following while loop
        # t_cl temperature of the outer surface of clothing
        t_cl = (r_a * t_skin + r_clo * t_op) / (r_a + r_clo)  # initial guess
        n_iterations = 0
        tc_converged = False

        while not tc_converged:

            # 0.95 is the clothing emissivity from ASHRAE fundamentals Ch. 9.7 Eq. 35
            if body_position == "sitting":
                # 0.7 ratio between radiation area of the body and the body area
                h_r = 4.0 * 0.95 * sbc * ((t_cl + tr) / 2.0 + 273.15) ** 3.0 * 0.7
            else:  # if standing
                # 0.73 ratio between radiation area of the body and the body area
                h_r = 4.0 * 0.95 * sbc * ((t_cl + tr) / 2.0 + 273.15) ** 3.0 * 0.73
            h_t = h_r + h_cc
            r_a = 1.0 / (f_a_cl * h_t)
            t_op = (h_r * tr + h_cc * tdb) / h_t
            t_cl_new = (r_a * t_skin + r_clo * t_op) / (r_a + r_clo)
            if abs(t_cl_new - t_cl) <= 0.01:
                tc_converged = True
            t_cl = t_cl_new
            n_iterations += 1

            if n_iterations > iteration_limit:
                raise StopIteration("Max iterations exceeded")

        q_sensible = (t_skin - t_op) / (r_a + r_clo)  # total sensible heat loss, W
        # hf_cs rate of energy transport between core and skin, W
        # 5.28 is the average body tissue conductance in W/(m2 C)
        # 1.163 is the thermal capacity of blood in Wh/(L C)
        hf_cs = (t_core - t_skin) * (5.28 + 1.163 * m_bl)
        s_core = m - hf_cs - q_res - c_res - wme  # rate of energy storage in the core
        s_skin = hf_cs - q_sensible - e_skin  # rate of energy storage in the skin
        tc_sk = 0.97 * alfa * body_weight  # thermal capacity skin
        tc_cr = 0.97 * (1 - alfa) * body_weight  # thermal capacity core
        d_t_sk = (s_skin * body_surface_area) / (
            tc_sk * 60.0
        )  # rate of change skin temperature °C per minute
        d_t_cr = (
            s_core * body_surface_area / (tc_cr * 60.0)
        )  # rate of change core temperature °C per minute
        t_skin = t_skin + d_t_sk
        t_core = t_core + d_t_cr
        t_body = alfa * t_skin + (1 - alfa) * t_core
        # sk_sig thermoregulatory control signal from the skin
        sk_sig = t_skin - temp_skin_neutral
        warm_sk = (sk_sig > 0) * sk_sig  # vasodilation signal
        colds = ((-1.0 * sk_sig) > 0) * (-1.0 * sk_sig)  # vasoconstriction signal
        # c_reg_sig thermoregulatory control signal from the skin, °C
        c_reg_sig = t_core - temp_core_neutral
        # c_warm vasodilation signal
        c_warm = (c_reg_sig > 0) * c_reg_sig
        # c_cold vasoconstriction signal
        c_cold = ((-1.0 * c_reg_sig) > 0) * (-1.0 * c_reg_sig)
        # bd_sig thermoregulatory control signal from the body
        bd_sig = t_body - temp_body_neutral
        warm_b = (bd_sig > 0) * bd_sig
        m_bl = (skin_blood_flow_neutral + c_dil * c_warm) / (1 + c_str * colds)
        if m_bl > max_skin_blood_flow:
            m_bl = max_skin_blood_flow
        if m_bl < 0.5:
            m_bl = 0.5
        m_rsw = c_sw * warm_b * math.exp(warm_sk / 10.7)  # regulatory sweating
        if m_rsw > max_sweating:
            m_rsw = max_sweating
        e_rsw = 0.68 * m_rsw  # heat lost by vaporization sweat
        r_ea = 1.0 / (lr * f_a_cl * h_cc)  # evaporative resistance air layer
        r_ecl = r_clo / (lr * i_cl)
        e_req = (
            rm - q_res - c_res - q_sensible
        )  # evaporative heat loss required for tmp regulation
        e_max = (math.exp(18.6686 - 4030.183 / (t_skin + 235.0)) - vapor_pressure) / (
            r_ea + r_ecl
        )
        if e_max == 0:  # added this otherwise e_rsw / e_max cannot be calculated
            e_max = 0.001
        p_rsw = e_rsw / e_max  # ratio heat loss sweating to max heat loss sweating
        w = 0.06 + 0.94 * p_rsw  # skin wetness
        e_diff = w * e_max - e_rsw  # vapor diffusion through skin
        if w > w_max:
            w = w_max
            p_rsw = w_max / 0.94
            e_rsw = p_rsw * e_max
            e_diff = 0.06 * (1.0 - p_rsw) * e_max
        if e_max < 0:
            e_diff = 0
            e_rsw = 0
            w = w_max
        e_skin = (
            e_rsw + e_diff
        )  # total evaporative heat loss sweating and vapor diffusion
        m_rsw = (
            e_rsw / 0.68
        )  # back calculating the mass of regulatory sweating as a function of e_rsw
        met_shivering = 19.4 * colds * c_cold  # met shivering W/m2
        m = rm + met_shivering
        alfa = 0.0417737 + 0.7451833 / (m_bl + 0.585417)

    q_skin = q_sensible + e_skin  # total heat loss from skin, W
    # p_s_sk saturation vapour pressure of water of the skin
    p_s_sk = math.exp(18.6686 - 4030.183 / (t_skin + 235.0))

    # standard environment - where _s at end of the variable names stands for standard
    h_r_s = h_r  # standard environment radiative heat transfer coefficient

    h_c_s = 3.0 * pow(pressure_in_atmospheres, 0.53)
    if not calculate_ce and met > 0.85:
        h_c_met = 5.66 * (met - 0.85) ** 0.39
        h_c_s = max(h_c_s, h_c_met)
    if h_c_s < 3.0:
        h_c_s = 3.0

    h_t_s = (
        h_c_s + h_r_s
    )  # sum of convective and radiant heat transfer coefficient W/(m2*K)
    r_clo_s = (
        1.52 / ((met - wme / met_factor) + 0.6944) - 0.1835
    )  # thermal resistance of clothing, °C M^2 /W
    r_cl_s = 0.155 * r_clo_s  # thermal insulation of the clothing in M2K/W
    f_a_cl_s = 1.0 + k_clo * r_clo_s  # increase in body surface area due to clothing
    f_cl_s = 1.0 / (
        1.0 + 0.155 * f_a_cl_s * h_t_s * r_clo_s
    )  # ratio of surface clothed body over nude body
    i_m_s = 0.45  # permeation efficiency of water vapour through the clothing layer
    i_cl_s = (
        i_m_s * h_c_s / h_t_s * (1 - f_cl_s) / (h_c_s / h_t_s - f_cl_s * i_m_s)
    )  # clothing vapor permeation efficiency
    r_a_s = 1.0 / (f_a_cl_s * h_t_s)  # resistance of air layer to dry heat
    r_ea_s = 1.0 / (lr * f_a_cl_s * h_c_s)
    r_ecl_s = r_cl_s / (lr * i_cl_s)
    h_d_s = 1.0 / (r_a_s + r_cl_s)
    h_e_s = 1.0 / (r_ea_s + r_ecl_s)

    # calculate Standard Effective Temperature (SET)
    delta = 0.0001
    dx = 100.0
    set_old = round(t_skin - q_skin / h_d_s, 2)
    while abs(dx) > 0.01:
        err_1 = (
            q_skin
            - h_d_s * (t_skin - set_old)
            - w
            * h_e_s
            * (p_s_sk - 0.5 * (math.exp(18.6686 - 4030.183 / (set_old + 235.0))))
        )
        err_2 = (
            q_skin
            - h_d_s * (t_skin - (set_old + delta))
            - w
            * h_e_s
            * (
                p_s_sk
                - 0.5 * (math.exp(18.6686 - 4030.183 / (set_old + delta + 235.0)))
            )
        )
        _set = set_old - delta * err_1 / (err_2 - err_1)
        dx = _set - set_old
        set_old = _set

    # calculate Effective Temperature (ET)
    h_d = 1 / (r_a + r_clo)
    h_e = 1 / (r_ea + r_ecl)
    et_old = t_skin - q_skin / h_d
    delta = 0.0001
    dx = 100.0
    while abs(dx) > 0.01:
        err_1 = (
            q_skin
            - h_d * (t_skin - et_old)
            - w
            * h_e
            * (p_s_sk - 0.5 * (math.exp(18.6686 - 4030.183 / (et_old + 235.0))))
        )
        err_2 = (
            q_skin
            - h_d * (t_skin - (et_old + delta))
            - w
            * h_e
            * (p_s_sk - 0.5 * (math.exp(18.6686 - 4030.183 / (et_old + delta + 235.0))))
        )
        et = et_old - delta * err_1 / (err_2 - err_1)
        dx = et - et_old
        et_old = et

    tbm_l = (0.194 / 58.15) * rm + 36.301  # lower limit for evaporative regulation
    tbm_h = (0.347 / 58.15) * rm + 36.669  # upper limit for evaporative regulation

    t_sens = 0.4685 * (t_body - tbm_l)  # predicted thermal sensation
    if (t_body >= tbm_l) & (t_body < tbm_h):
        t_sens = w_max * 4.7 * (t_body - tbm_l) / (tbm_h - tbm_l)
    elif t_body >= tbm_h:
        t_sens = w_max * 4.7 + 0.4685 * (t_body - tbm_h)

    disc = (
        4.7 * (e_rsw - e_comfort) / (e_max * w_max - e_comfort - e_diff)
    )  # predicted thermal discomfort
    if disc <= 0:
        disc = t_sens

    # PMV Gagge
    pmv_gagge = (0.303 * math.exp(-0.036 * m) + 0.028) * (e_req - e_comfort - e_diff)

    # PMV SET
    dry_set = h_d_s * (t_skin - _set)
    e_req_set = rm - c_res - q_res - dry_set
    pmv_set = (0.303 * math.exp(-0.036 * m) + 0.028) * (e_req_set - e_comfort - e_diff)

    # Predicted  Percent  Satisfied  With  the  Level  of  Air  Movement
    ps = 100 * (1.13 * (t_op**0.5) - 0.24 * t_op + 2.7 * (v**0.5) - 0.99 * v)

    return (
        _set,
        e_skin,
        e_rsw,
        e_max,
        q_sensible,
        q_skin,
        q_res,
        t_core,
        t_skin,
        m_bl,
        m_rsw,
        w,
        w_max,
        et,
        pmv_gagge,
        pmv_set,
        disc,
        t_sens,
    )


@vectorize(
    [
        float64(
            float64,
            float64,
            float64,
            float64,
            float64,
            float64,
            float64,
            float64,
            float64,
            float64,
        )
    ],
)
def _two_nodes_optimized_return_set(
    tdb,
    tr,
    v,
    met,
    clo,
    vapor_pressure,
    wme,
    body_surface_area,
    p_atmospheric,
    body_position,
):
    return _two_nodes_optimized(
        tdb,
        tr,
        v,
        met,
        clo,
        vapor_pressure,
        wme,
        body_surface_area,
        p_atmospheric,
        body_position,
        True,
    )[0]



from pythermalcomfort.models import zhang_sensation_comfort
from pprint import pprint

# -------------------------------------------
# This model is often used with a thermal physiology model.
# However, this function can be used when you have physiological data and wants to use Zhang's comfort model.
# For example, this function can be used when you want to predict thermal sensation or comfort
# but you only have skin temperature or core temperature data measured by a subject experiment
# or when you have skin temperature data from a thermal manikin".
#
# If you have a time series of skin temperature data, calculate and enter the time derivative of skin temperature.
# If you want to do a steady state simulation, change 0 to it.
# The time derivative of core temperature is set to be 0 because it does change due to human thermoregulation
# except in an extremely hot or cold environment.
# -------------------------------------------

dict_results = zhang_sensation_comfort(
    t_skin_local={
        "head": 34.3,
        "neck": 34.6,
        "chest": 34.1,
        "back": 34.3,
        "pelvis": 34.3,
        "left_shoulder": 33.2,
        "left_arm": 33.6,
        "left_hand": 33.4,
        "right_shoulder": 33.2,
        "right_arm": 33.6,
        "right_hand": 33.4,
        "left_thigh": 33.3,
        "left_leg": 31.8,
        "left_foot": 32.3,
        "right_thigh": 33.3,
        "right_leg": 31.8,
        "right_foot": 32.3,
    },
    dt_skin_local_dt={
        "head": 0.01,
        "neck": 0.01,
        "chest": 0.01,
        "back": 0.01,
        "pelvis": 0.01,
        "left_shoulder": 0.01,
        "left_arm": 0.01,
        "left_hand": 0.01,
        "right_shoulder": 0.01,
        "right_arm": 0.01,
        "right_hand": 0.01,
        "left_thigh": 0.01,
        "left_leg": 0.01,
        "left_foot": 0.01,
        "right_thigh": 0.01,
        "right_leg": 0.01,
        "right_foot": 0.01,
    },
    dt_core_local_dt=0,
    options={
        "sensation_scale": "7-point",
    },
)

pprint(dict_results)



from typing import Union, List

import numpy as np

from pythermalcomfort.utilities import (
    units_converter,
)


def clo_tout(
    tout: Union[float, int, np.ndarray, List[float], List[int]], units: str = "SI"
) -> Union[float, np.ndarray]:
    """Representative clothing insulation Icl as a function of outdoor air
    temperature at 06:00 a.m [4]_.

    Parameters
    ----------
    tout : float, int, or array-like
        outdoor air temperature at 06:00 a.m., default in [°C] in [°F] if `units` = 'IP'
    units : str, optional
        select the SI (International System of Units) or the IP (Imperial Units) system.
        Supported values are 'SI' and 'IP'. Defaults to 'SI'.

    Returns
    -------
    clo : float, int, or array-like
         Representative clothing insulation Icl, [clo]

    Raises
    -------
    TypeError
        If `tout` is not a float, int, NumPy array, or a list of floats or integers.
    ValueError
        If an invalid unit is provided or non-numeric elements are found in `tout`.

    Notes
    -----
    The ASHRAE 55 2020 states that it is acceptable to determine the clothing
    insulation Icl using this equation in mechanically conditioned buildings [1]_.

    Limitations:
        - This equation may not be accurate for extreme temperature ranges.

    Examples
    --------
    .. code-block:: python

        >>> from pythermalcomfort.models import clo_tout
        >>> clo_tout(tout=27)  # Can be int or float
        0.46
        >>> clo_tout(tout=[27, 25])  # List of ints or floats
        array([0.46, 0.47])
    """

    # Use explicit type hints for sequence
    valid_types = (
        float,
        int,
        np.ndarray,
        list,
    )  # Use 'list' instead of 'List[float]' or 'List[int]'
    if not isinstance(tout, valid_types):
        raise TypeError("tout must be a float, int, NumPy array, or a list.")

    # Check for valid list elements separately
    if isinstance(tout, list):
        if not all(isinstance(item, (float, int)) for item in tout):
            raise TypeError("Elements of tout list must be floats or integers.")

    # Convert tout to NumPy array for vectorized operations
    tout = np.array(tout)

    # Validate units string
    valid_units: List[str] = ["SI", "IP"]
    if units.upper() not in valid_units:
        raise ValueError(f"Invalid unit: {units}. Supported units are {valid_units}.")

    # Convert units if necessary
    if units.lower() == "ip":
        tout = units_converter(tmp=tout)[0]

    clo = np.where(tout < 26, np.power(10, -0.1635 - 0.0066 * tout), 0.46)
    clo = np.where(tout < 5, 0.818 - 0.0364 * tout, clo)
    clo = np.where(tout < -5, 1, clo)

    return np.around(clo, 2)



import unittest
import numpy as np
from pythermalcomfort.models import clo_tout


class TestCloTout(unittest.TestCase):
    def test_si_units_scalar(self):
        # Test SI units with a scalar outdoor air temperature
        result = clo_tout(tout=27)
        self.assertEqual(result, 0.46)

    def test_si_units_array(self):
        # Test SI units with an array of outdoor air temperatures
        result = clo_tout(tout=[27, 25])
        np.testing.assert_equal(result, [0.46, 0.47])

    def test_ip_units_scalar(self):
        # Test IP units with a scalar outdoor air temperature
        result = clo_tout(tout=80.6, units="IP")
        self.assertEqual(result, 0.46)

    def test_ip_units_array(self):
        # Test IP units with an array of outdoor air temperatures
        result = clo_tout(tout=[80.6, 77], units="IP")
        np.testing.assert_equal(result, np.array([0.46, 0.47]))

    def test_edge_cases(self):
        # Test edge cases of the piecewise function
        result = clo_tout(tout=[4, 2, 0])
        np.testing.assert_equal(result, [0.67, 0.75, 0.82])

        result = clo_tout(tout=[-4, -6])
        np.testing.assert_equal(result, [0.96, 1.0])

    def test_invalid_units(self):
        # Test invalid units
        with self.assertRaises(ValueError):
            clo_tout(tout=27, units="invalid")

    def test_invalid_type(self):
        # Test invalid type for tout
        with self.assertRaises(TypeError):
            clo_tout(tout="invalid")



import numpy as np

from pythermalcomfort.models.two_nodes import two_nodes
from pythermalcomfort.utilities import (
    units_converter,
    check_standard_compliance_array,
    body_surface_area,
)


def set_tmp(
    tdb,
    tr,
    v,
    rh,
    met,
    clo,
    wme=0,
    body_surface_area=1.8258,
    p_atm=101325,
    body_position="standing",
    units="SI",
    limit_inputs=True,
    **kwargs,
):
    """
    Calculates the Standard Effective Temperature (SET). The SET is the temperature of
    a hypothetical isothermal environment at 50% (rh), <0.1 m/s (20 fpm) average air
    speed (v), and tr = tdb, in which the total heat loss from the skin of an imaginary occupant
    wearing clothing, standardized for the activity concerned is the same as that
    from a person in the actual environment with actual clothing and activity level. [10]_

    Parameters
    ----------
    tdb : float or array-like
        dry bulb air temperature, default in [°C] in [°F] if `units` = 'IP'
    tr : float or array-like
        mean radiant temperature, default in [°C] in [°F] if `units` = 'IP'
    v : float or array-like
        air speed, default in [m/s] in [fps] if `units` = 'IP'
    rh : float or array-like
        relative humidity, [%]
    met : float or array-like
        metabolic rate, [met]
    clo : float or array-like
        clothing insulation, [clo]
    wme : float or array-like
        external work, [met] default 0
    body_surface_area : float
        body surface area, default value 1.8258 [m2] in [ft2] if `units` = 'IP'

        The body surface area can be calculated using the function
        :py:meth:`pythermalcomfort.utilities.body_surface_area`.
    p_atm : float
        atmospheric pressure, default value 101325 [Pa] in [atm] if `units` = 'IP'
    body_position: str default="standing" or array-like
        select either "sitting" or "standing"
    units : {'SI', 'IP'}
        select the SI (International System of Units) or the IP (Imperial Units) system.
    limit_inputs : boolean default True
        By default, if the inputs are outsude the following limits the
        function returns nan. If False returns values regardless of the input values.
        The limits are 10 < tdb [°C] < 40, 10 < tr [°C] < 40,
        0 < v [m/s] < 2, 1 < met [met] < 4, and 0 < clo [clo] < 1.5.

    Other Parameters
    ----------------
    round : boolean, deafult True
        if True rounds output value, if False it does not round it

    Returns
    -------
    SET : float or array-like
        Standard effective temperature, [°C]

    Notes
    -----
    You can use this function to calculate the `SET`_ temperature in accordance with
    the ASHRAE 55 2020 Standard [1]_.

    .. _SET: https://en.wikipedia.org/wiki/Thermal_comfort#Standard_effective_temperature

    Examples
    --------
    .. code-block:: python

        >>> from pythermalcomfort.models import set_tmp
        >>> set_tmp(tdb=25, tr=25, v=0.1, rh=50, met=1.2, clo=.5)
        24.3
        >>> set_tmp(tdb=[25, 25], tr=25, v=0.1, rh=50, met=1.2, clo=.5)
        array([24.3, 24.3])

        >>> # for users who wants to use the IP system
        >>> set_tmp(tdb=77, tr=77, v=0.328, rh=50, met=1.2, clo=.5, units='IP')
        75.8

    """
    # When SET is used to calculate CE then h_c is calculated in a slightly different way
    default_kwargs = {"round": True, "calculate_ce": False}
    kwargs = {**default_kwargs, **kwargs}

    if units.lower() == "ip":
        if body_surface_area == 1.8258:
            body_surface_area = 19.65
        if p_atm == 101325:
            p_atm = 1
        tdb, tr, v, body_surface_area, p_atm = units_converter(
            tdb=tdb, tr=tr, v=v, area=body_surface_area, pressure=p_atm
        )

    tdb = np.array(tdb)
    tr = np.array(tr)
    v = np.array(v)
    rh = np.array(rh)
    met = np.array(met)
    clo = np.array(clo)
    wme = np.array(wme)

    set_array = two_nodes(
        tdb=tdb,
        tr=tr,
        v=v,
        rh=rh,
        met=met,
        clo=clo,
        wme=wme,
        body_surface_area=body_surface_area,
        p_atmospheric=p_atm,
        body_position=body_position,
        calculate_ce=kwargs["calculate_ce"],
        round=False,
        output="all",
    )["_set"]

    if units.lower() == "ip":
        set_array = units_converter(tmp=set_array, from_units="si")[0]

    if limit_inputs:
        (
            tdb_valid,
            tr_valid,
            v_valid,
            met_valid,
            clo_valid,
        ) = check_standard_compliance_array(
            "ashrae", tdb=tdb, tr=tr, v=v, met=met, clo=clo
        )
        all_valid = ~(
            np.isnan(tdb_valid)
            | np.isnan(tr_valid)
            | np.isnan(v_valid)
            | np.isnan(met_valid)
            | np.isnan(clo_valid)
        )
        set_array = np.where(all_valid, set_array, np.nan)

    if kwargs["round"]:
        return np.around(set_array, 1)
    else:
        return set_array



import numpy as np
from pythermalcomfort.jos3_functions.parameters import Default
from pythermalcomfort.utilities import DefaultSkinTemperature

# Coefficients for sensation prediction
c1_plus = {
    "head": 0.9,
    "neck": 0.9,
    "chest": 1.0,
    "back": 1.0,
    "pelvis": 0.4,
    "left_shoulder": 0.4,
    "left_arm": 0.7,
    "left_hand": 0.45,
    "right_shoulder": 0.4,
    "right_arm": 0.7,
    "right_hand": 0.45,
    "left_thigh": 0.29,
    "left_leg": 0.4,
    "left_foot": 0.26,
    "right_thigh": 0.29,
    "right_leg": 0.4,
    "right_foot": 0.26,
}

c1_minus = {
    "head": 0.38,
    "neck": 0.38,
    "chest": 0.35,
    "back": 0.3,
    "pelvis": 0.2,
    "left_shoulder": 0.29,
    "left_arm": 0.3,
    "left_hand": 0.2,
    "right_shoulder": 0.29,
    "right_arm": 0.3,
    "right_hand": 0.2,
    "left_thigh": 0.2,
    "left_leg": 0.29,
    "left_foot": 0.25,
    "right_thigh": 0.2,
    "right_leg": 0.29,
    "right_foot": 0.25,
}

k1 = {
    "head": 0.18,
    "neck": 0.18,
    "chest": 0.1,
    "back": 0.1,
    "pelvis": 0.15,
    "left_shoulder": 0.1,
    "left_arm": 0.1,
    "left_hand": 0.15,
    "right_shoulder": 0.1,
    "right_arm": 0.1,
    "right_hand": 0.15,
    "left_thigh": 0.11,
    "left_leg": 0.1,
    "left_foot": 0.15,
    "right_thigh": 0.11,
    "right_leg": 0.1,
    "right_foot": 0.15,
}

c2_plus = {
    "head": 90,
    "neck": 90,
    "chest": 136,
    "back": 192,
    "pelvis": 137,
    "left_shoulder": 167,
    "left_arm": 125,
    "left_hand": 46,
    "right_shoulder": 167,
    "right_arm": 125,
    "right_hand": 46,
    "left_thigh": 263,
    "left_leg": 212,
    "left_foot": 162,
    "right_thigh": 263,
    "right_leg": 212,
    "right_foot": 162,
}

c2_minus = {
    "head": 543,
    "neck": 543,
    "chest": 39,
    "back": 88,
    "pelvis": 75,
    "left_shoulder": 156,
    "left_arm": 144,
    "left_hand": 19,
    "right_shoulder": 156,
    "right_arm": 144,
    "right_hand": 19,
    "left_thigh": 151,
    "left_leg": 206,
    "left_foot": 109,
    "right_thigh": 151,
    "right_leg": 206,
    "right_foot": 109,
}

c3 = {
    "head": 0,
    "neck": 0,
    "chest": -2135,
    "back": -4054,
    "pelvis": -5053,
    "left_shoulder": 0,
    "left_arm": 0,
    "left_hand": 0,
    "right_shoulder": 0,
    "right_arm": 0,
    "right_hand": 0,
    "left_thigh": 0,
    "left_leg": 0,
    "left_foot": 0,
    "right_thigh": 0,
    "right_leg": 0,
    "right_foot": 0,
}

a_if_delta_sens_local_is_less_than_minus_2 = {
    "head": 0.54,
    "neck": 0.65,
    "chest": 0.91,
    "back": 0.91,
    "pelvis": 0.94,
    "left_shoulder": 0.43,
    "left_arm": 0.37,
    "left_hand": 0.25,
    "right_shoulder": 0.43,
    "right_arm": 0.37,
    "right_hand": 0.25,
    "left_thigh": 0.81,
    "left_leg": 0.7,
    "left_foot": 0.5,
    "right_thigh": 0.81,
    "right_leg": 0.7,
    "right_foot": 0.5,
}

a_if_delta_sens_local_is_between_minus_2_to_plus_2 = {
    "head": 0.5,
    "neck": 0.46,
    "chest": 0.57,
    "back": 0.46,
    "pelvis": 0.32,
    "left_shoulder": 0.28,
    "left_arm": 0.38,
    "left_hand": 0.0,
    "right_shoulder": 0.28,
    "right_arm": 0.38,
    "right_hand": 0.0,
    "left_thigh": 0.3,
    "left_leg": 0.29,
    "left_foot": 0.0,
    "right_thigh": 0.3,
    "right_leg": 0.29,
    "right_foot": 0.0,
}

a_if_delta_sens_local_is_plus_2_or_more = {
    "head": 0.4,
    "neck": 0.4,
    "chest": 0.4,
    "back": 0.2,
    "pelvis": 0.6,
    "left_shoulder": 0.4,
    "left_arm": 0.3,
    "left_hand": 0.1,
    "right_shoulder": 0.4,
    "right_arm": 0.3,
    "right_hand": 0.1,
    "left_thigh": 0.4,
    "left_leg": 0.4,
    "left_foot": 0.6,
    "right_thigh": 0.4,
    "right_leg": 0.4,
    "right_foot": 0.6,
}

b_if_delta_sens_local_is_less_than_minus_2 = {
    "head": -1.1,
    "neck": -0.92,
    "chest": -1.14,
    "back": -0.92,
    "pelvis": -0.64,
    "left_shoulder": -0.56,
    "left_arm": -0.73,
    "left_hand": 0.0,
    "right_shoulder": -0.56,
    "right_arm": -0.73,
    "right_hand": 0.0,
    "left_thigh": -0.6,
    "left_leg": -0.59,
    "left_foot": 0.0,
    "right_thigh": -0.6,
    "right_leg": -0.59,
    "right_foot": 0.0,
}

b_if_delta_sens_local_is_between_minus_2_to_plus_2 = {
    "head": 0.0,
    "neck": 0.0,
    "chest": 0.0,
    "back": 0.0,
    "pelvis": 0.0,
    "left_shoulder": 0.0,
    "left_arm": 0.0,
    "left_hand": 0.0,
    "right_shoulder": 0.0,
    "right_arm": 0.0,
    "right_hand": 0.0,
    "left_thigh": 0.0,
    "left_leg": 0.0,
    "left_foot": 0.0,
    "right_thigh": 0.0,
    "right_leg": 0.0,
    "right_foot": 0.0,
}

b_if_delta_sens_local_is_plus_2_or_more = {
    "head": 1.1,
    "neck": 0.63,
    "chest": 1.14,
    "back": 0.92,
    "pelvis": 0.64,
    "left_shoulder": 0.56,
    "left_arm": 0.73,
    "left_hand": 0.0,
    "right_shoulder": 0.56,
    "right_arm": 0.73,
    "right_hand": 0.0,
    "left_thigh": 0.6,
    "left_leg": 0.59,
    "left_foot": 0.0,
    "right_thigh": 0.6,
    "right_leg": 0.59,
    "right_foot": 0.0,
}

c_if_delta_sens_local_is_less_than_minus_2 = {
    "head": -2,
    "neck": -2,
    "chest": -2,
    "back": -2,
    "pelvis": -2,
    "left_shoulder": -2,
    "left_arm": -2,
    "left_hand": -2,
    "right_shoulder": -2,
    "right_arm": -2,
    "right_hand": -2,
    "left_thigh": -2,
    "left_leg": -2,
    "left_foot": -2,
    "right_thigh": -2,
    "right_leg": -2,
    "right_foot": -2,
}

c_if_delta_sens_local_is_between_minus_2_to_plus_2 = {
    "head": 0.0,
    "neck": 0.0,
    "chest": 0.0,
    "back": 0.0,
    "pelvis": 0.0,
    "left_shoulder": 0.0,
    "left_arm": 0.0,
    "left_hand": 0.0,
    "right_shoulder": 0.0,
    "right_arm": 0.0,
    "right_hand": 0.0,
    "left_thigh": 0.0,
    "left_leg": 0.0,
    "left_foot": 0.0,
    "right_thigh": 0.0,
    "right_leg": 0.0,
    "right_foot": 0.0,
}

c_if_delta_sens_local_is_plus_2_or_more = {
    "head": 2,
    "neck": 2,
    "chest": 2,
    "back": 2,
    "pelvis": 2,
    "left_shoulder": 2,
    "left_arm": 2,
    "left_hand": 2,
    "right_shoulder": 2,
    "right_arm": 2,
    "right_hand": 2,
    "left_thigh": 2,
    "left_leg": 2,
    "left_foot": 2,
    "right_thigh": 2,
    "right_leg": 2,
    "right_foot": 2,
}

# Coefficients for comfort prediction
C31 = {
    "head": -0.35,
    "neck": 0.0,
    "chest": -0.66,
    "back": -0.45,
    "pelvis": -0.59,
    "left_shoulder": -0.3,
    "left_arm": -0.23,
    "left_hand": -0.8,
    "right_shoulder": -0.3,
    "right_arm": -0.23,
    "right_hand": -0.8,
    "left_thigh": 0.0,
    "left_leg": -0.2,
    "left_foot": -0.91,
    "right_thigh": 0.0,
    "right_leg": -0.2,
    "right_foot": -0.91,
}

C32 = {
    "head": 0.35,
    "neck": 0.0,
    "chest": 0.66,
    "back": 0.45,
    "pelvis": 0.0,
    "left_shoulder": 0.35,
    "left_arm": 0.23,
    "left_hand": 0.8,
    "right_shoulder": 0.35,
    "right_arm": 0.23,
    "right_hand": 0.8,
    "left_thigh": 0.0,
    "left_leg": 0.61,
    "left_foot": 0.4,
    "right_thigh": 0.0,
    "right_leg": 0.61,
    "right_foot": 0.4,
}

C6 = {
    "head": 2.17,
    "neck": 1.96,
    "chest": 2.1,
    "back": 2.1,
    "pelvis": 2.06,
    "left_shoulder": 2.14,
    "left_arm": 2.0,
    "left_hand": 1.98,
    "right_shoulder": 2.14,
    "right_arm": 2.0,
    "right_hand": 1.98,
    "left_thigh": 1.98,
    "left_leg": 2.0,
    "left_foot": 2.13,
    "right_thigh": 1.98,
    "right_leg": 2.0,
    "right_foot": 2.13,
}

C71 = {
    "head": 0.28,
    "neck": 0.0,
    "chest": 1.39,
    "back": 0.96,
    "pelvis": 0.5,
    "left_shoulder": 0.0,
    "left_arm": 0.0,
    "left_hand": 0.48,
    "right_shoulder": 0.0,
    "right_arm": 0.0,
    "right_hand": 0.48,
    "left_thigh": 0.0,
    "left_leg": 1.67,
    "left_foot": 0.5,
    "right_thigh": 0.0,
    "right_leg": 1.67,
    "right_foot": 0.5,
}

C72 = {
    "head": 0.4,
    "neck": 0.0,
    "chest": 0.9,
    "back": 0.0,
    "pelvis": 0.0,
    "left_shoulder": 0.0,
    "left_arm": 1.71,
    "left_hand": 0.48,
    "right_shoulder": 0.0,
    "right_arm": 1.71,
    "right_hand": 0.48,
    "left_thigh": 0.0,
    "left_leg": 0.0,
    "left_foot": 0.3,
    "right_thigh": 0.0,
    "right_leg": 0.0,
    "right_foot": 0.3,
}

C8 = {
    "head": 0.5,
    "neck": -0.19,
    "chest": 0.0,
    "back": 0.0,
    "pelvis": -0.51,
    "left_shoulder": -0.4,
    "left_arm": -0.68,
    "left_hand": 0.0,
    "right_shoulder": -0.4,
    "right_arm": -0.68,
    "right_hand": 0.0,
    "left_thigh": 0.0,
    "left_leg": 0.0,
    "left_foot": 0.0,
    "right_thigh": 0.0,
    "right_leg": 0.0,
    "right_foot": 0.0,
}
c1_plus_array = np.array(list(c1_plus.values()))
c1_minus_array = np.array(list(c1_minus.values()))
k1_array = np.array(list(k1.values()))
c2_plus_array = np.array(list(c2_plus.values()))
c2_minus_array = np.array(list(c2_minus.values()))
c3_array = np.array(list(c3.values()))
a_if_delta_sens_local_is_less_than_minus_2_array = np.array(
    list(a_if_delta_sens_local_is_less_than_minus_2.values())
)
a_if_delta_sens_local_is_between_minus_2_to_plus_2_array = np.array(
    list(a_if_delta_sens_local_is_between_minus_2_to_plus_2.values())
)
a_if_delta_sens_local_is_plus_2_or_more_array = np.array(
    list(a_if_delta_sens_local_is_plus_2_or_more.values())
)
b_if_delta_sens_local_is_less_than_minus_2_array = np.array(
    list(b_if_delta_sens_local_is_less_than_minus_2.values())
)
b_if_delta_sens_local_is_between_minus_2_to_plus_2_array = np.array(
    list(b_if_delta_sens_local_is_between_minus_2_to_plus_2.values())
)
b_if_delta_sens_local_is_plus_2_or_more_array = np.array(
    list(b_if_delta_sens_local_is_plus_2_or_more.values())
)
c_if_delta_sens_local_is_less_than_minus_2_array = np.array(
    list(c_if_delta_sens_local_is_less_than_minus_2.values())
)
c_if_delta_sens_local_is_between_minus_2_to_plus_2_array = np.array(
    list(c_if_delta_sens_local_is_between_minus_2_to_plus_2.values())
)
c_if_delta_sens_local_is_plus_2_or_more_array = np.array(
    list(c_if_delta_sens_local_is_plus_2_or_more.values())
)
C31_array = np.array(list(C31.values()))
C32_array = np.array(list(C32.values()))
C6_array = np.array(list(C6.values()))
C71_array = np.array(list(C71.values()))
C72_array = np.array(list(C72.values()))
C8_array = np.array(list(C8.values()))

# 17 body segment names (The name corresponds to JOS-3 model)
body_name_list = [
    "head",
    "neck",
    "chest",
    "back",
    "pelvis",
    "left_shoulder",
    "left_arm",
    "left_hand",
    "right_shoulder",
    "right_arm",
    "right_hand",
    "left_thigh",
    "left_leg",
    "left_foot",
    "right_thigh",
    "right_leg",
    "right_foot",
]


def convert_to_17_segments_array_from_any_data_type(data):
    """Convert various input data types to a 17-segment numpy array.

    This function is designed to handle a variety of input types including integers, floats, dictionaries,
    lists, and numpy arrays, and returns a standardized 17-segment numpy array. This can be useful when
    ensuring consistent array formats across different input types.

    Parameters
    ----------
    data : int, float, dict, list, ndarray or None
        Input data that needs to be converted. If `data` is:
        - int or float: The returned array will have all 17 values set to this number.
        - dict: The keys of the dictionary should match the `body_name_list` and the returned array will be constructed based on this order.
        - list or ndarray: Must be of length 17.

    Returns
    -------
    data_17_array : numpy.ndarray
        A numpy array of shape (17,) containing the standardized data.

    Raises
    ------
    ValueError
        If the input data is not one of the supported types, or if it is a list or ndarray of length other than 17.

    Notes
    -----
    Ensure that the `body_name_list` is defined elsewhere in the code if using this function with dictionary input data.
    """
    if data is None:
        return None
    # Convert input data to sens_third_coldest numpy array for consistent handling
    if isinstance(data, (int, float)):
        data_17_array = np.ones(17) * data
    elif isinstance(data, dict):
        data_17_array = np.array([data[key] for key in body_name_list])
    elif isinstance(data, (list, np.ndarray)):
        data_17_array = np.asarray(data)
        if data.shape == (17,):
            data_17_array = data
        else:
            ValueError("The input list or ndarray is not of length 17")
    else:
        raise ValueError(
            "Unsupported input type. Supported types: int, float, list, dict, ndarray"
        )
    return data_17_array


def convert_17_segments_dict_from_array(array, body_name_list):
    """
    Convert a 17-segment numpy array to a dictionary using the provided body name list.

    Parameters
    ----------
    array : numpy.ndarray
        A numpy array of shape (17,) containing the data to be converted.

    body_name_list : list
        A list of body part names with length 17. Each name corresponds to a value in the `array`.

    Returns
    -------
    dict
        A dictionary with keys as body part names from `body_name_list` and values from the input `array`.
    """
    return {part: array[i] for i, part in enumerate(body_name_list)}


def calculate_local_thermal_sensation(
    t_skin_local_array,
    dt_skin_local_dt_array,
    t_skin_local_set_array,
    t_skin_mean_value,
    t_skin_mean_set_value,
    dt_core_local_dt_array,
    sensation_scale,
):
    """Calculate local thermal sensation based on physiological parameters and the selected sensation scale.

    Parameters
    -----------
    t_skin_local_array : numpy.ndarray
        Array of local skin temperatures [°C].
    dt_skin_local_dt_array : numpy.ndarray
        Time derivative of local skin temperatures [°C/s].
    t_skin_local_set_array : numpy.ndarray
        Array of local skin temperature setpoints [°C].
    t_skin_mean_value : float
        Mean skin temperature value [°C].
    t_skin_mean_set_value : float
        Setpoint for mean skin temperature [°C].
    dt_core_local_dt_array : numpy.ndarray
        Time derivative of core temperatures [°C/s].
    sensation_scale : str
        Thermal sensation scale to be used. Accepts either '9-point' or '7-point'. Default is "9-point".
        - '9-point': Ranges from -4 (very cold) to 4 (very hot).
        - '7-point': Ranges from -3 (cold) to 3 (hot).

    Return
    ------
    sensation_local_array : numpy.ndarray
        Local thermal sensation on sens_third_coldest 9-point/7-point sensation scale [-].
    """

    def get_coefficient_c1(t_skin_local_array, t_skin_local_set_array):
        """Return c1_minus or c1_plus based on the comparison of skin temperature to its setpoint."""
        return np.where(
            t_skin_local_array - t_skin_local_set_array < 0,
            c1_minus_array,
            c1_plus_array,
        )

    def get_coefficient_c2(dt_skin_local_dt_array):
        """Return c2_minus or c2_plus based on the local skin temperature derivative."""
        return np.where(dt_skin_local_dt_array < 0, c2_minus_array, c2_plus_array)

    # Set coefficients for the equation
    c1 = get_coefficient_c1(
        t_skin_local_array=t_skin_local_array,
        t_skin_local_set_array=t_skin_local_set_array,
    )

    k1 = k1_array
    c2 = get_coefficient_c2(dt_skin_local_dt_array=dt_skin_local_dt_array)

    c3 = c3_array

    # Calculate static component of local thermal sensation
    sensation_local_static_array = 4 * (
        (
            2
            / (
                1
                + np.exp(
                    -c1 * (t_skin_local_array - t_skin_local_set_array)
                    - k1
                    * (
                        (t_skin_local_array - t_skin_local_set_array)
                        - (t_skin_mean_value - t_skin_mean_set_value)
                    )
                )
            )
        )
        - 1
    )

    # Calculate dynamic component of local thermal sensation
    sensation_local_dynamic_array = (
        c2 * dt_skin_local_dt_array + c3 * dt_core_local_dt_array
    )

    # Calculate local thermal sensation
    sensation_local_array = sensation_local_static_array + sensation_local_dynamic_array

    # Apply the selected thermal sensation scale
    if sensation_scale == "7-point":
        # Clip the array between -3 and 3
        sensation_local_array = np.clip(sensation_local_array, -3, 3)
    elif sensation_scale == "9-point":
        # Clip the array between -4 and 4
        sensation_local_array = np.clip(sensation_local_array, -4, 4)
    else:
        raise ValueError("Please select '7-point' or '9-point'")

    return sensation_local_array


def calculate_overall_sensation(sensation_local_array):
    """Calculate overall thermal sensation based on local sensation for each body part.

    Sensations in the cold or warm range are emphasized more in the average.
    If the most extreme sensations are from the hands or feet, they are excluded from the weighted average.

    Parameters
    ----------
    sensation_local_array : ndarray
        Array or list of local thermal sensations from various body parts.

    Returns
    -------
    overall_sensation : float
        Overall thermal sensation.
    """

    # Define local thermal comfort as a dictionary
    local_17_segments_dict = convert_17_segments_dict_from_array(
        array=sensation_local_array, body_name_list=body_name_list
    )

    def are_most_extreme_sensations_hands_or_feet(
        sensation_local_array,
        left_hand_sensation,
        right_hand_sensation,
        left_foot_sensation,
        right_foot_sensation,
    ):
        """Determine if the most extreme sensations in a given array are from the hands or feet.

        This function examines the input array of sensations to determine if the most
        extreme sensations (both maximum and minimum) come from the hands or feet. The
        sensations of the hands and feet are provided separately for comparison.

        Parameters
        ----------
        sensation_local_array : list or ndarray
            An array or list containing sensations from various body parts.
        left_hand_sensation : float
            The sensation value corresponding to the left hand.
        right_hand_sensation : float
            The sensation value corresponding to the right hand.
        left_foot_sensation : float
            The sensation value corresponding to the left foot.
        right_foot_sensation : float
            The sensation value corresponding to the right foot.

        Returns
        -------
        bool
            Returns True if the most extreme sensations (either the two largest or two smallest)
            correspond to both hands or both feet. Otherwise, returns False.

        Notes
        -----
        The function assumes that the `sensation_local_array` contains at least two elements.
        If not, it will directly return False without further checks.

        """

        # Sensation array must have at least two elements for this function to work properly
        if len(sensation_local_array) < 2:
            return False

        # Extract the top two maximum and minimum sensations
        max_two = sorted(sensation_local_array, reverse=True)[:2]
        min_two = sorted(sensation_local_array)[:2]

        # Set up hand and foot sensations
        hand_sensations = {left_hand_sensation, right_hand_sensation}
        foot_sensations = {left_foot_sensation, right_foot_sensation}

        # Check if the most extreme sensations correspond to two hands or two feet
        return (set(max_two) == hand_sensations or set(max_two) == foot_sensations) or (
            set(min_two) == hand_sensations or set(min_two) == foot_sensations
        )

    def calculate_weighted_overall_sensation(
        sensation_local_array,
        cold_threshold=-2,
        warm_threshold=2,
        coefficient=2,
    ):
        """Calculate the weighted average of provided thermal sensations.

        This function calculates a weighted average of given local thermal sensations, adjusting
        weights dynamically based on the provided thresholds. Sensations that fall within the
        cold or warm threshold range are given increased importance in the average using
        a specified coefficient.

        If the most extreme sensations are from the hands or feet, they are excluded from the weighted average.

        Parameters
        ----------
        sensation_local_array : list or ndarray
            An array or list containing thermal sensations for various body parts.
        cold_threshold : float, optional
            The threshold below which sensations are considered cold and given higher weight.
            Default is -2.
        warm_threshold : float, optional
            The threshold above which sensations are considered warm and given higher weight.
            Default is 2.
        coefficient : float, optional
            The coefficient to increase the weight of sensations that fall within the
            cold or warm threshold. Default is 2.

        Returns
        -------
        overall_sensation_weighted : float
            The weighted overall sensation.
        """

        # Calculate the absolute values of sensations as initial weights
        abs_sensations = np.abs(sensation_local_array)

        # Check if most extreme sensations are hands or feet
        if are_most_extreme_sensations_hands_or_feet(
            sensation_local_array=sensation_local_array,
            left_hand_sensation=local_17_segments_dict["left_hand"],
            right_hand_sensation=local_17_segments_dict["right_hand"],
            left_foot_sensation=local_17_segments_dict["left_foot"],
            right_foot_sensation=local_17_segments_dict["right_foot"],
        ):
            # Set the weight of the sensation with the maximum absolute value to 0
            max_abs_index = np.argmax(np.abs(sensation_local_array))
            abs_sensations[max_abs_index] = 0

        # Apply the coefficient to sensations that fall within the specified ranges
        abs_sensations[sensation_local_array <= cold_threshold] *= coefficient
        abs_sensations[sensation_local_array >= warm_threshold] *= coefficient

        # Normalize the weights so that their sum is 1
        total_weight = np.sum(abs_sensations)
        normalized_weights = (
            abs_sensations / total_weight if total_weight != 0 else abs_sensations
        )

        # Calculate the weighted average
        overall_sensation_weighted = np.average(
            sensation_local_array, weights=normalized_weights
        )

        return overall_sensation_weighted

    overall_sensation = calculate_weighted_overall_sensation(
        sensation_local_array, cold_threshold=-2, warm_threshold=2, coefficient=2
    )

    return overall_sensation


def calculate_local_thermal_comfort(overall_sensation, sensation_local_array):
    """Calculate local thermal comfort based on overall sensation and local sensations.

    This function computes the local thermal comfort by considering the overall sensation and
    an array of local sensations. It utilizes predefined coefficients and applies specific formulas
    based on whether the overall sensation is positive or negative.

    Parameters
    ----------
    overall_sensation : float
        The overall sensation which can be a positive or negative float.
    sensation_local_array : np.ndarray
        A numpy array containing local sensations.

    Returns
    -------
    local_comfort_array : np.ndarray
        local thermal comfort values. Each value is constrained to the range of -4 (very uncomfortable) to 4 (very comfortable).

    """

    C31 = C31_array
    C32 = C32_array
    C6 = C6_array
    C71 = C71_array
    C72 = C72_array
    C8 = C8_array

    def get_elements_for_calculating_local_thermal_comfort(overall_sensation):
        """Calculate and return elements used for determining local thermal comfort.

        Parameters
        ----------
        overall_sensation : float
            The overall thermal sensation which could be either positive or negative.

        Returns
        -------
        tuple
            A tuple containing the computed elements.
        """
        if overall_sensation >= 0:
            A1 = -8 * (C6 + C72 * overall_sensation + 4)
            A2 = (
                (4 - (C8 + C32 * overall_sensation))
                * (-4 - (C8 + C32 * overall_sensation))
                * (1 + np.exp(5 * (sensation_local_array + C32 * overall_sensation)))
            )
            A4 = (4 + (C6 + C72 * overall_sensation)) / (
                -4 - (C8 + C32 * overall_sensation)
            )
            A5 = sensation_local_array + C8 + C32 * overall_sensation
            A6 = C6 + C72 * overall_sensation
        else:
            A1 = -8 * (C6 - C71 * overall_sensation + 4)
            A2 = (
                (4 - (C8 - C31 * overall_sensation))
                * (-4 - (C8 - C31 * overall_sensation))
                * (1 + np.exp(5 * (sensation_local_array - C31 * overall_sensation)))
            )
            A4 = (4 + (C6 - C71 * overall_sensation)) / (
                -4 - (C8 - C31 * overall_sensation)
            )
            A5 = sensation_local_array + C8 - C31 * overall_sensation
            A6 = C6 - C71 * overall_sensation

        return A1, A2, A4, A5, A6

    sensation_local_array = np.array(sensation_local_array)
    local_comfort_array = np.zeros(len(sensation_local_array))

    for i in range(len(sensation_local_array)):
        A1, A2, A4, A5, A6 = get_elements_for_calculating_local_thermal_comfort(
            overall_sensation
        )

        local_comfort_array = (A1 / A2 + A4) * A5 + A6

        # Constrain local_comfort_values[i] to the range -4 to 4
        if local_comfort_array[i] < -4:
            local_comfort_array[i] = -4
        elif local_comfort_array[i] > 4:
            local_comfort_array[i] = 4

    return local_comfort_array


def calculate_overall_thermal_comfort(
    local_comfort_array, is_transient=False, is_controlled=False
):
    """Calculate the overall thermal comfort based on local comfort values and additional conditions.

    This function calculates the overall thermal comfort by considering local thermal comforts
    and their relation to the hands and feet.

    It uses two rules:
    1) When in steady state or controlled environments, and
    2) In transient or uncontrolled conditions.
    The function adjusts the weights of the local comfort values based on these rules.

    Parameters
    ----------
    local_comfort_array : list or ndarray
        An array or list containing local thermal comfort values for various body segments.
    is_transient : bool, optional
        A flag indicating if the environment is in a transient state. Default is False.

        Note : The way to determine whether transient or not is not described in the paper.
        Original CBE's C++ code always returns False (It might be determined by the time derivative of skin temperature and so on).

    is_controlled : bool, optional
        A flag indicating if the environment is controlled. Default is False.
        This can be changed to "True" if simulated people can adjust their own thermal environment,
        such as in an automobile or personal comfort system.

    Returns
    -------
    overall_comfort_value : float
        The overall thermal comfort value.

    Notes
    -----
    - Rule 1: In steady state or controlled environments, the overall comfort is the average of
      the two lowest local comfort votes unless the two lowest sensations are from hands or feet.
    - Rule 2: In transient or uncontrolled conditions, the overall comfort is the average of the
      two lowest and the highest comfort votes, unless the two lowest sensations are from hands or feet.
    - The function makes use of the `are_lowest_comfort_values_hands_or_feet` inner function to
      determine if the two most uncomfortable segments are the hands or feet.
    - The function assumes that `body_name_list` and `convert_17_segments_dict_from_array` are available
      in the surrounding code.
    """

    # Define local comfort dictionary
    local_comfort_dict = convert_17_segments_dict_from_array(
        array=local_comfort_array, body_name_list=body_name_list
    )

    # Define the highest local thermal comfort (most comfortable)
    highest_local_comfort_value = np.max(local_comfort_array)

    # Sort in ascending order
    local_comfort_ascending_order_array = np.sort(local_comfort_array)

    # Define the TOP-4 lowest local thermal comfort (TOP-4 most uncomfortable)
    (
        lowest_local_comfort_value,
        second_lowest_local_comfort_value,
        third_lowest_local_comfort_value,
        fourth_lowest_local_comfort_value,
    ) = local_comfort_ascending_order_array[:4]

    # Define the mean value of TOP-2 lowest local thermal comfort
    mean_top2_lowest_comfort = np.mean(
        [lowest_local_comfort_value, second_lowest_local_comfort_value]
    )

    def are_lowest_comfort_values_hands_or_feet(local_comfort_dict):
        """
        Determine if the two most uncomfortable sensations are from the hands or feet.

        Parameters
        ----------
        local_comfort_dict : dict
            A dictionary containing body segments as keys and their associated thermal comfort values.
            It should at least have entries for "left_hand", "right_hand", "left_foot", and "right_foot".

        Returns
        -------
        bool
            True if the two most uncomfortable sensations correspond to the hands or feet. False otherwise.
        """

        # Find the two minimum comfort values
        two_lowest_values = sorted(local_comfort_dict.values())[:2]

        # Extract comfort values for hands and feet
        comfort_left_hand = local_comfort_dict["left_hand"]
        comfort_right_hand = local_comfort_dict["right_hand"]
        comfort_left_foot = local_comfort_dict["left_foot"]
        comfort_right_foot = local_comfort_dict["right_foot"]

        # Compare the two lowest values with the comfort values for hands and feet
        # If the two lowest values match the values for either hands or feet, return True
        return (
            sorted([comfort_left_hand, comfort_right_hand]) == two_lowest_values
            or sorted([comfort_left_foot, comfort_right_foot]) == two_lowest_values
        )

    # Rule 1 (Steady state or controlled, such as in sens_third_coldest climate chamber)
    # Rule 1: Overall comfort is the average of the two minimum local comfort votes unless Rule 2 applies.
    if not is_transient or is_controlled:
        if are_lowest_comfort_values_hands_or_feet == True:
            overall_comfort_value = np.mean(
                [
                    mean_top2_lowest_comfort,
                    third_lowest_local_comfort_value,
                    fourth_lowest_local_comfort_value,
                ]
            )
        else:  # False
            overall_comfort_value = np.mean(
                [
                    lowest_local_comfort_value,
                    second_lowest_local_comfort_value,
                    third_lowest_local_comfort_value,
                ]
            )

    # Rule 2 (Transient state or not controlled space)
    # Rule 2: Overall comfort is the average of the two minimum votes and the maximum comfort vote:
    else:
        if are_lowest_comfort_values_hands_or_feet == True:
            overall_comfort_value = np.mean(
                [
                    mean_top2_lowest_comfort,
                    third_lowest_local_comfort_value,
                    highest_local_comfort_value,
                ]
            )
        else:  # False
            overall_comfort_value = np.mean(
                [
                    lowest_local_comfort_value,
                    second_lowest_local_comfort_value,
                    highest_local_comfort_value,
                ]
            )

    return overall_comfort_value


def zhang_sensation_comfort(
    t_skin_local,
    dt_skin_local_dt,
    dt_core_local_dt,
    t_skin_local_set=DefaultSkinTemperature()._asdict(),
    options=None,
):
    """Zhang's local and overall sensation and comfort model.

    This model predicts thermal sensation and comfort based on physiological responses,
    such as skin temperature and its time variation.

    It consists of four main helper models that predict:
    1. Local thermal sensation
    2. Whole-body thermal sensation
    3. Local thermal comfort
    4. Whole-body thermal comfort

    The results of these predictions are returned as a dictionary.

    Accepted input data types for conditions like skin temperature are int, float, list, dict, and ndarray.
    To avoid input errors, it's recommended to provide the data as a dictionary.
    Each key in the dictionary should represent the name of a body segment, aligning with the JOS-3 model.

    body_name_list = [
        "head", "neck", "chest", "back", "pelvis",
        "left_shoulder", "left_arm", "left_hand",
        "right_shoulder", "right_arm", "right_hand",
        "left_thigh", "left_leg", "left_foot",
        "right_thigh", "right_leg", "right_foot"
    ]

    When data is provided as an int or float format, all body segments will be assigned the same value.
    If data is provided as a list or ndarray format, it should be of length 17, with values corresponding to the keys listed above.


    Parameters
    -----------
    t_skin_local : int, float, list, numpy.ndarray, dict
        Local skin temperature [°C].
    t_skin_local_set : int, float, list, numpy.ndarray, dict
        Local skin temperature setpoint [°C].

        Note: t_skin_local_set is the local skin temperature when people feel thermally neutral.
        Neutral skin temperatures vary by site, generally higher for the trunk and lower for the extremities.
        It is stated that multi-node segment model is used to get these parameters under the condition where
        people thermally feel neutral, calculated by the PMV model at a certain metabolic rate and clothing level.

        The experimental data on neutral skin temperature from Zhang's experiments : https://escholarship.org/uc/item/3f4599hx

    dt_skin_local_dt : int, float, list, numpy.ndarray, dict
        Time derivetive of local skin temperature [°C/s].
    dt_core_local_dt : int, float, list, numpy.ndarray, dict
        Time derivetive of core temperature [°C/s].
    options : dict, optional
        Dictionary containing options for the calculation.

        Keys include:
            "sensation_scale" (str):
                Scale for evaluating thermal sensation. Default is "9-point".
                Either '9-point' or '7-point'.
                9-point sensation scale (-4:very cold, -3:cold, -2:cool, -1:slightly cool, 0:neutral, +1:slightly warm, +2:warm, +3:hot, +4:very hot)
                7-point sensation scale (-3:cold, -2:cool, -1:slightly cool, 0:neutral, +1:slightly warm, +2:warm, +3:hot)

            "is_transient" (bool):
                Indicates if the environment is transient. The original paper does not describe how to define.
                Default is False.

            "is_controlled" (bool):
                Indicates if the environment is controlled. Default is False.

            "show_extra_info" (bool):
                Indicates if extra information should be displayed. Default is False.

    Returns
    -------
    sensation_local : local thermal sensation (each body segment) [-]
    sensation_overall : overall thermal sensation [-]
    comfort_local : local thermal comfort for each body segment [-]
    comfort_overall : overall thermal sensation [-]
    OPTIONAL PARAMETERS : the paramters listed below are returned if show_extra_info = True
    t_skin_local : local skin temperature [°C].
    t_skin_local_set : local skin temperature setpoint [°C].
    dt_skin_local_dt : time derivetive of local skin temperature [°C/s].
    t_skin_mean : mean skin temperature for the whole body [°C].
    t_skin_mean_set : mean skin temperature setpoint for the whole body [°C].
    dt_core_local_dt : time derivetive of core temperature [°C/s].


    Examples
    ---------
    .. code-block:: python
    >>> from pythermalcomfort.models import zhang_sensation_comfort
    >>> dict_results = zhang_sensation_comfort(
    >>>     t_skin_local={
    >>>         "head": 34.3,
    >>>         "neck": 34.6,
    >>>         "chest": 34.1,
    >>>         "back": 34.3,
    >>>         "pelvis": 34.3,
    >>>         "left_shoulder": 33.2,
    >>>         "left_arm": 33.6,
    >>>         "left_hand": 33.4,
    >>>         "right_shoulder": 33.2,
    >>>         "right_arm": 33.6,
    >>>         "right_hand": 33.4,
    >>>         "left_thigh": 33.3,
    >>>         "left_leg": 31.8,
    >>>         "left_foot": 32.3,
    >>>         "right_thigh": 33.3,
    >>>         "right_leg": 31.8,
    >>>         "right_foot": 32.3,
    >>>     },
    >>>     dt_skin_local_dt={
    >>>         "head": 0.01,
    >>>         "neck": 0.01,
    >>>         "chest": 0.01,
    >>>         "back": 0.01,
    >>>         "pelvis": 0.01,
    >>>         "left_shoulder": 0.01,
    >>>         "left_arm": 0.01,
    >>>         "left_hand": 0.01,
    >>>         "right_shoulder": 0.01,
    >>>         "right_arm": 0.01,
    >>>         "right_hand": 0.01,
    >>>         "left_thigh": 0.01,
    >>>         "left_leg": 0.01,
    >>>         "left_foot": 0.01,
    >>>         "right_thigh": 0.01,
    >>>         "right_leg": 0.01,
    >>>         "right_foot": 0.01,
    >>>     },
    >>>     dt_core_local_dt=0,
    >>>     options={
    >>>         "sensation_scale": "7-point",
    >>>     },
    >>> )
    >>>
    >>> pprint(dict_results)
    {
    'comfort_back': -0.48,
    'comfort_chest': 0.98,
    'comfort_head': 1.35,
    'comfort_left_arm': 3.74,
    'comfort_left_foot': 0.13,
    'comfort_left_hand': 1.12,
    'comfort_left_leg': -0.97,
    'comfort_left_shoulder': 0.36,
    'comfort_left_thigh': -1.36,
    'comfort_neck': 1.98,
    'comfort_overall': -1.23,
    'comfort_pelvis': 1.27,
    'comfort_right_arm': 3.74,
    'comfort_right_foot': 0.13,
    'comfort_right_hand': 1.12,
    'comfort_right_leg': -0.97,
    'comfort_right_shoulder': 0.36,
    'comfort_right_thigh': -1.36,
    'sensation_back': 1.32,
    'sensation_chest': 0.67,
    'sensation_head': 0.15,
    'sensation_left_arm': 0.65,
    'sensation_left_foot': 1.12,
    'sensation_left_hand': 0.06,
    'sensation_left_leg': 1.54,
    'sensation_left_shoulder': 1.09,
    'sensation_left_thigh': 2.23,
    'sensation_neck': 0.15,
    'sensation_overall': 1.42,
    'sensation_pelvis': 0.97,
    'sensation_right_arm': 0.65,
    'sensation_right_foot': 1.12,
    'sensation_right_hand': 0.06,
    'sensation_right_leg': 1.54,
    'sensation_right_shoulder': 1.09,
    'sensation_right_thigh': 2.23
     }
    """

    default_options = {
        "sensation_scale": "9-point",
        "is_transient": False,
        "is_controlled": False,
        "show_extra_info": False,
    }

    # If options is provided, update the default options with the provided ones
    if options:
        default_options.update(options)

    options = default_options  # Now use the merged dictionary in your function

    # Define phygiological input parameters
    t_skin_local_array = convert_to_17_segments_array_from_any_data_type(t_skin_local)
    dt_skin_local_dt_array = convert_to_17_segments_array_from_any_data_type(
        dt_skin_local_dt
    )
    t_skin_local_set_array = convert_to_17_segments_array_from_any_data_type(
        t_skin_local_set
    )
    t_skin_mean_value = np.average(t_skin_local_array, weights=Default.local_bsa)
    t_skin_mean_set_value = np.average(
        t_skin_local_set_array, weights=Default.local_bsa
    )
    dt_core_local_dt_array = convert_to_17_segments_array_from_any_data_type(
        dt_core_local_dt
    )

    # Define optional input parameters
    sensation_scale = options["sensation_scale"]
    is_transient = options["is_transient"]
    is_controlled = options["is_controlled"]
    show_extra_info = options["show_extra_info"]

    # ------------------------------------------------------
    # Process Main
    # ------------------------------------------------------

    # Calculate local thermal sensation
    sensation_local_array = calculate_local_thermal_sensation(
        t_skin_local_array=t_skin_local_array,
        dt_skin_local_dt_array=dt_skin_local_dt_array,
        t_skin_local_set_array=t_skin_local_set_array,
        t_skin_mean_value=t_skin_mean_value,
        t_skin_mean_set_value=t_skin_mean_set_value,
        dt_core_local_dt_array=dt_core_local_dt_array,
        sensation_scale=sensation_scale,
    )

    # Calculate overall thermal sensation
    overall_sensation_value = calculate_overall_sensation(
        sensation_local_array=sensation_local_array
    )

    # Calculate local thermal comfort
    comfort_local_array = calculate_local_thermal_comfort(
        overall_sensation=overall_sensation_value,
        sensation_local_array=sensation_local_array,
    )

    # Calculate overall thermal comfort
    overall_comfort_value = calculate_overall_thermal_comfort(
        local_comfort_array=comfort_local_array,
        is_transient=is_transient,
        is_controlled=is_controlled,
    )

    # Define result dict
    dict_data = {
        "sensation_local": np.around(sensation_local_array, 2),
        "sensation_overall": np.around(overall_sensation_value, 2),
        "comfort_local": np.around(comfort_local_array, 2),
        "comfort_overall": np.around(overall_comfort_value, 2),
    }

    if show_extra_info:
        dict_data.update(
            {
                "t_skin_local": t_skin_local_array,
                "dt_skin_local_dt": dt_skin_local_dt_array,
                "t_skin_local_set": t_skin_local_set_array,
                "t_skin_mean_value": t_skin_mean_value,
                "t_skin_mean_set_value": t_skin_mean_set_value,
                "dt_core_local_dt_array": dt_core_local_dt_array,
                "sensation_scale": sensation_scale,
                "is_transient": is_transient,
            }
        )

    # Create new dictionary to store the new keys and values
    new_data_dict = {}

    # Map 17 arrays of values to new keys for keys that contain 'local'
    for key in dict_data.keys():
        if "local" in key:
            # Map 17 arrays of values to new keys
            for i, body_name in enumerate(body_name_list):
                # Replace 'local' with body_name in the key
                new_key = key.replace("local", body_name)
                new_value = dict_data[key][i]
                new_data_dict[new_key] = new_value

    # Merge the new dictionary with the original dictionary
    dict_data.update(new_data_dict)

    # Delete the original 'local_sensation' and 'local_comfort' keys
    for key in list(dict_data.keys()):
        if "local" in key and not any(part in key for part in body_name_list):
            del dict_data[key]

    return dict_data



from pythermalcomfort.models import humidex
import pytest


class TestHumidex:

    #  Calculate humidex for a given dry bulb air temperature and relative humidity
    def test_calculate_humidex(self):
        result = humidex(25, 50)
        assert result["humidex"] == pytest.approx(28.2, abs=0.1)

    #  Return the humidex and discomfort level for a given dry bulb air temperature and relative humidity
    def test_return_humidex_and_discomfort(self):
        result = humidex(25, 50)
        assert "humidex" in result
        assert "discomfort" in result

    #  Round the output humidex value if the 'round' parameter is set to True
    def test_round_output_humidex(self):
        result = humidex(25, 50, round=True)
        assert isinstance(result["humidex"], int) or isinstance(
            result["humidex"], float
        )

    #  Return an error if the input dry bulb air temperature is not a float
    def test_input_temperature_not_float(self):
        with pytest.raises(TypeError):
            humidex("25", 50)

    #  Return an error if the input relative humidity is not a float
    def test_input_humidity_not_float(self):
        with pytest.raises(TypeError):
            humidex(25, "50")

    #  Return "Evident discomfort" if the calculated humidex is between 35 and 40
    def test_humidex_evident_discomfort(self):
        result = humidex(10, 50)
        assert result["discomfort"] == "Little or no discomfort"

        result = humidex(28, 50)
        assert result["discomfort"] == "Noticeable discomfort"

        result = humidex(30, 50)
        assert result["discomfort"] == "Evident discomfort"

        result = humidex(35, 50)
        assert result["discomfort"] == "Intense discomfort; avoid exertion"

        result = humidex(40, 50)
        assert result["discomfort"] == "Heat stroke probable"

    #  Return an error if the input relative humidity is greater than 100%
    def test_humidex_invalid_rh(self):
        with pytest.raises(ValueError):
            humidex(tdb=25, rh=110)

    #  Return an error if the input relative humidity is less than 0%
    def test_input_relative_humidity_less_than_zero(self):
        with pytest.raises(ValueError):
            humidex(25, -10)

    #  Return "Little or no discomfort" if the calculated humidex is exactly 30
    def test_humidex_discomfort_30(self):
        result = humidex(25, 50)
        assert result["discomfort"] == "Little or no discomfort"

    #  Return "Dangerous discomfort" if the calculated humidex is exactly 54
    def test_humidex_dangerous_discomfort(self):
        result = humidex(30, 100)
        assert result["discomfort"] == "Dangerous discomfort"

    def test_humidex(self):
        assert humidex(25, 50) == {
            "humidex": 28.2,
            "discomfort": "Little or no discomfort",
        }
        assert humidex(30, 80) == {
            "humidex": 43.3,
            "discomfort": "Intense discomfort; avoid exertion",
        }
        assert humidex(31.6, 57.1) == {
            "humidex": 40.8,
            "discomfort": "Intense discomfort; avoid exertion",
        }



def wbgt(twb, tg, tdb=None, with_solar_load=False, **kwargs):
    """Calculates the Wet Bulb Globe Temperature (WBGT) index calculated in
    compliance with the ISO 7243 [11]_. The WBGT is a heat stress index that
    measures the thermal environment to which a person is exposed. In most
    situations, this index is simple to calculate. It should be used as a
    screening tool to determine whether heat stress is present. The PHS model
    allows a more accurate estimation of stress. PHS can be calculated using
    the function :py:meth:`pythermalcomfort.models.phs`.

    The WBGT determines the impact of heat on a person throughout the course of a working
    day (up to 8 h). It does not apply to very brief heat exposures. It pertains to
    the evaluation of male and female people who are fit for work in both indoor
    and outdoor occupational environments, as well as other sorts of surroundings [11]_.

    The WBGT is defined as a function of only twb and tg if the person is not exposed to
    direct radiant heat from the sun. When a person is exposed to direct radiant heat,
    tdb must also be specified.

    Parameters
    ----------
    twb : float,
        natural (no forced air flow) wet bulb temperature, [°C]
    tg : float
        globe temperature, [°C]
    tdb : float
        dry bulb air temperature, [°C]. This value is needed as input if the person is
        exposed to direct solar radiation
    with_solar_load: bool
        True if the globe sensor is exposed to direct solar radiation

    Other Parameters
    ----------------
    round: boolean, default True
        if True rounds output value, if False it does not round it

    Returns
    -------
    wbgt : float
        Wet Bulb Globe Temperature Index, [°C]

    Examples
    --------
    .. code-block:: python

        >>> from pythermalcomfort.models import wbgt
        >>> wbgt(twb=25, tg=32)
        27.1

        >>> # if the persion is exposed to direct solar radiation
        >>> wbgt(twb=25, tg=32, tdb=20, with_solar_load=True)
        25.9
    """
    default_kwargs = {
        "round": True,
    }
    kwargs = {**default_kwargs, **kwargs}

    if with_solar_load and tdb is None:
        raise ValueError("Please enter the dry bulb air temperature")

    if with_solar_load:
        t_wbg = 0.7 * twb + 0.2 * tg + 0.1 * tdb
    else:
        t_wbg = 0.7 * twb + 0.3 * tg

    if kwargs["round"]:
        return round(t_wbg, 1)
    else:
        return t_wbg



import pytest

from pythermalcomfort.models import pet_steady


def test_pet():
    assert pet_steady(tdb=20, tr=20, rh=50, v=0.15, met=1.37, clo=0.5) == pytest.approx(
        18.85, abs=0.1
    )
    assert pet_steady(tdb=30, tr=30, rh=50, v=0.15, met=1.37, clo=0.5) == pytest.approx(
        30.6, abs=0.01
    )
    assert pet_steady(tdb=20, tr=20, rh=50, v=0.5, met=1.37, clo=0.5) == pytest.approx(
        17.16, abs=0.01
    )
    assert pet_steady(tdb=21, tr=21, rh=50, v=0.1, met=1.37, clo=0.9) == pytest.approx(
        21.08, abs=0.01
    )
    assert pet_steady(tdb=20, tr=20, rh=50, v=0.1, met=1.37, clo=0.9) == pytest.approx(
        19.92, abs=0.01
    )
    assert pet_steady(tdb=-5, tr=40, rh=2, v=0.5, met=1.37, clo=0.9) == pytest.approx(
        7.82, abs=0.01
    )
    assert pet_steady(tdb=-5, tr=-5, rh=50, v=5.0, met=1.37, clo=0.9) == pytest.approx(
        -13.38, abs=0.01
    )
    assert pet_steady(tdb=30, tr=60, rh=80, v=1.0, met=1.37, clo=0.9) == pytest.approx(
        44.63, abs=0.01
    )
    assert pet_steady(tdb=30, tr=30, rh=80, v=1.0, met=1.37, clo=0.9) == pytest.approx(
        32.21, abs=0.01
    )



import pytest
import math

from pythermalcomfort.models import wbgt


class TestWbgt:
    def test_wbgt(self):
        assert math.isclose(wbgt(twb=25, tg=32, tdb=20, with_solar_load=True), 25.9)
        with pytest.raises(ValueError):
            wbgt(twb=25, tg=32, with_solar_load=True)
        # data from Table D.1 ISO 7243
        assert math.isclose(wbgt(twb=17.3, tg=40, round=True), 24.1)
        assert math.isclose(wbgt(twb=21.1, tg=55, round=True), 31.3)
        assert math.isclose(wbgt(twb=16.7, tg=40, round=True), 23.7)

        #  Calculate WBGT with twb and tg

    def test_calculate_wbgt_with_twb_and_tg(self):
        assert math.isclose(wbgt(25, 30), 26.5)

    #  Calculate WBGT with twb, tg, and tdb
    def test_calculate_wbgt_with_twb_tg_and_tdb(self):
        assert math.isclose(wbgt(25, 30, 20), 26.5)

    #  Round WBGT to one decimal place
    def test_round_wbgt_to_one_decimal_place(self):
        assert math.isclose(wbgt(25, 30, round=False), 26.5)

    #  Calculate WBGT with twb and tg set to 0
    def test_calculate_wbgt_with_twb_and_tg_set_to_zero(self):
        assert math.isclose(wbgt(0, 0), 0.0)

    #  Calculate WBGT with tdb set to 0
    def test_calculate_wbgt_with_tdb_set_to_zero(self):
        assert math.isclose(wbgt(25, 30, 0), 26.5)

    #  Calculate WBGT with twb and tg set to None
    def test_calculate_wbgt_with_twb_and_tg_set_to_none(self):
        with pytest.raises(TypeError):
            wbgt(None, None)



import math
from typing import List, Union

import numpy as np
from numba import jit

from pythermalcomfort.psychrometrics import p_sat
from pythermalcomfort.utilities import (
    check_standard_compliance_array,
)


def phs(
    tdb: Union[float, int, np.ndarray, List[float], List[int]],
    tr: Union[float, int, np.ndarray, List[float], List[int]],
    v: Union[float, int, np.ndarray, List[float], List[int]],
    rh: Union[float, int, np.ndarray, List[float], List[int]],
    met: Union[float, int, np.ndarray, List[float], List[int]],
    clo: Union[float, int, np.ndarray, List[float], List[int]],
    posture,
    wme: Union[float, int, np.ndarray, List[float], List[int]] = 0,
    **kwargs
):
    """Calculates the Predicted Heat Strain (PHS) index based in compliace with
    the ISO 7933:2004 Standard [8]_. The ISO 7933 provides a method for the
    analytical evaluation and interpretation of the thermal stress experienced
    by a subject in a hot environment. It describes a method for predicting the
    sweat rate and the internal core temperature that the human body will
    develop in response to the working conditions.

    The PHS model can be used to predict the: heat by respiratory convection, heat flow
    by respiratory evaporation, steady state mean skin temperature, instantaneous value
    of skin temperature, heat accumulation associated with the metabolic rate, maximum
    evaporative heat flow at the skin surface, predicted sweat rate, predicted evaporative
    heat flow, and rectal temperature.

    Parameters
    ----------
    tdb : float, int, or array-like
        dry bulb air temperature, default in [°C]
    tr : float, int, or array-like
        mean radiant temperature, default in [°C]
    v : float, int, or array-like
        air speed, default in [m/s]
    rh : float, int, or array-like
        relative humidity, [%]
    met : float, int, or array-like
        metabolic rate, [W/(m2)]
    clo : float, int, or array-like
        clothing insulation, [clo]
    posture: int
        a numeric value presenting posture of person [sitting=1, standing=2, crouching=3]
    wme : float, int, or array-like
        external work, [W/(m2)] default 0

    Other Parameters
    ----------------
    limit_inputs : boolean default True
        By default, if the inputs are outsude the standard applicability limits the
        function returns nan. If False returns values even if input values are
        outside the applicability limits of the model.

        The 7933 limits are 15 < tdb [°C] < 50, 0 < tr [°C] < 60,
        0 < vr [m/s] < 3, 100 < met [met] < 450, and 0.1 < clo [clo] < 1.
    i_mst : float, default 0.38
        static moisture permeability index, [dimensionless]
    a_p : float, default 0.54
        fraction of the body surface covered by the reflective clothing, [dimensionless]
    drink : int, default 1
        1 if workers can drink freely, 0 otherwise
    weight : float, default 75
        body weight, [kg]
    height : float, default 1.8
        height, [m]
    walk_sp : float, default 0
        walking speed, [m/s]
    theta : float, default 0
        angle between walking direction and wind direction [degrees]
    acclimatized : int, default 100
        100 if acclimatized subject, 0 otherwise
    duration : int, default 480
        duration of the work sequence, [minutes]
    f_r : float, default 0.97
        emissivity of the reflective clothing, [dimensionless]
        Some reference values :py:meth:`pythermalcomfort.utilities.f_r_garments`.
    t_sk : float, default 34.1
        mean skin temperature when worker starts working, [°C]
    t_cr : float, default 36.8
        mean core temperature when worker starts working, [°C]
    t_re : float, default False
        mean rectal temperature when worker starts working, [°C]
    t_cr_eq : float, default False
        mean core temperature as a function of met when worker starts working, [°C]
    sweat_rate : float, default 0

    Returns
    -------
    t_re : float
        rectal temperature, [°C]
    t_sk : float
        skin temperature, [°C]
    t_cr : float
        core temperature, [°C]
    t_cr_eq : float
        core temperature as a function of the metabolic rate, [°C]
    t_sk_t_cr_wg : float
        fraction of the body mass at the skin temperature
    d_lim_loss_50 : float
        maximum allowable exposure time for water loss, mean subject, [minutes]
    d_lim_loss_95 : float
        maximum allowable exposure time for water loss, 95% of the working population,
        [minutes]
    d_lim_t_re : float
        maximum allowable exposure time for heat storage, [minutes]
    water_loss_watt : float
        maximum water loss in watts, [W]
    water_loss : float
        maximum water loss, [g]

    Examples
    --------
    .. code-block:: python

        >>> from pythermalcomfort.models import phs
        >>> results = phs(tdb=40, tr=40, rh=33.85, v=0.3, met=150, clo=0.5, posture=2, wme=0)
        >>> print(results)
        {'t_re': 37.5, 'd_lim_loss_50': 440, 'd_lim_loss_95': 298, 'd_lim_t_re': 480,
        'water_loss': 6166.0}
    """
    default_kwargs = {
        "i_mst": 0.38,
        "a_p": 0.54,
        "drink": 1,
        "weight": 75,
        "height": 1.8,
        "walk_sp": 0,
        "theta": 0,
        "acclimatized": 100,
        "duration": 480,
        "f_r": 0.97,
        "t_sk": 34.1,
        "t_cr": 36.8,
        "t_re": False,
        "t_cr_eq": False,
        "t_sk_t_cr_wg": 0.3,
        "sweat_rate": 0,
        "round": True,
        "limit_inputs": True,
    }
    kwargs = {**default_kwargs, **kwargs}

    tdb = np.array(tdb)
    tr = np.array(tr)
    v = np.array(v)
    rh = np.array(rh)
    met = np.array(met)
    clo = np.array(clo)
    wme = np.array(wme)
    posture = np.array(posture)

    i_mst = kwargs["i_mst"]
    a_p = kwargs["a_p"]
    drink = kwargs["drink"]
    weight = kwargs["weight"]
    height = kwargs["height"]
    walk_sp = kwargs["walk_sp"]
    theta = kwargs["theta"]
    acclimatized = kwargs["acclimatized"]
    duration = kwargs["duration"]
    f_r = kwargs["f_r"]
    t_sk = kwargs["t_sk"]
    t_cr = kwargs["t_cr"]
    t_re = kwargs["t_re"]
    t_cr_eq = kwargs["t_cr_eq"]
    t_sk_t_cr_wg = kwargs["t_sk_t_cr_wg"]
    sweat_rate = kwargs["sweat_rate"]
    limit_inputs = kwargs["limit_inputs"]

    p_a = p_sat(tdb) / 1000 * rh / 100

    acclimatized = int(acclimatized)
    if acclimatized < 0 or acclimatized > 100:
        raise ValueError("Acclimatized should be between 0 and 100")

    if drink not in [0, 1]:
        raise ValueError("Drink should be 0 or 1")

    if weight <= 0 or weight > 1000:
        raise ValueError(
            "The weight of the person should be in kg and it cannot exceed 1000"
        )

    # I needed to create this variable since vectorize accept no more than 34 inputs+outputs
    combined_drink_acc_weight = drink * 1000000 + acclimatized * 1000 + weight

    if not t_re:
        t_re = t_cr
    if not t_cr_eq:
        t_cr_eq = t_cr

    (
        t_re,
        t_sk,
        t_cr,
        t_cr_eq,
        t_sk_t_cr_wg,
        sweat_rate,
        sw_tot_g,
        d_lim_loss_50,
        d_lim_loss_95,
        d_lim_t_re,
    ) = np.vectorize(_phs_optimized, cache=True)(
        tdb=tdb,
        tr=tr,
        v=v,
        p_a=p_a,
        met=met,
        clo=clo,
        posture=posture,
        combined_drink_acc_weight=combined_drink_acc_weight,
        wme=wme,
        i_mst=i_mst,
        a_p=a_p,
        height=height,
        walk_sp=walk_sp,
        theta=theta,
        duration=duration,
        f_r=f_r,
        t_sk=t_sk,
        t_cr=t_cr,
        t_re=t_re,
        t_cr_eq=t_cr_eq,
        t_sk_t_cr_wg=t_sk_t_cr_wg,
        sw_tot=sweat_rate,
    )

    output = {
        "t_re": t_re,
        "t_sk": t_sk,
        "t_cr": t_cr,
        "t_cr_eq": t_cr_eq,
        "t_sk_t_cr_wg": t_sk_t_cr_wg,
        "d_lim_loss_50": d_lim_loss_50,
        "d_lim_loss_95": d_lim_loss_95,
        "d_lim_t_re": d_lim_t_re,
        "water_loss_watt": sweat_rate,
        "water_loss": sw_tot_g,
    }

    if limit_inputs:
        (
            tdb_valid,
            tr_valid,
            v_valid,
            p_a_valid,
            met_valid,
            clo_valid,
        ) = check_standard_compliance_array(
            "7933", tdb=tdb, tr=tr, v=v, met=met, clo=clo, p_a=p_a
        )
        all_valid = ~(
            np.isnan(tdb_valid)
            | np.isnan(tr_valid)
            | np.isnan(v_valid)
            | np.isnan(p_a_valid)
            | np.isnan(met_valid)
            | np.isnan(clo_valid)
        )
        for key in output.keys():
            output[key] = np.where(all_valid, output[key], np.nan)

    if kwargs["round"]:
        for key in output.keys():
            if key != "t_sk_t_cr_wg":
                output[key] = np.around(output[key], 1)
            else:
                output[key] = np.around(output[key], 2)

    return output


# Constants
const_t_eq = math.exp(-1 / 10)
const_t_sk = math.exp(-1 / 3)
const_sw = math.exp(-1 / 10)


@jit(nopython=True)
def _phs_optimized(
    tdb,
    tr,
    v,
    p_a,
    met,
    clo,
    posture,
    combined_drink_acc_weight,
    wme,
    i_mst,
    a_p,
    height,
    walk_sp,
    theta,
    duration,
    f_r,
    t_sk,
    t_cr,
    t_re,
    t_cr_eq,
    t_sk_t_cr_wg,
    sw_tot,
):
    drink = int(combined_drink_acc_weight / 1000000)
    acclimatized = int((combined_drink_acc_weight - drink * 1000000) / 1000)
    weight = combined_drink_acc_weight - drink * 1000000 - acclimatized * 1000

    # DuBois body surface area [m2]
    a_dubois = 0.202 * (weight**0.425) * (height**0.725)
    sp_heat = 57.83 * weight / a_dubois  # specific heat of the body
    d_lim_t_re = 0  # maximum allowable exposure time for heat storage [min]
    # maximum allowable exposure time for water loss, mean subject [min]
    d_lim_loss_50 = 0
    # maximum allowable exposure time for water loss, 95 % of the working population [min]
    d_lim_loss_95 = 0
    # maximum water loss to protect a mean subject [g]
    d_max_50 = 0.075 * weight * 1000
    # maximum water loss to protect 95 % of the working population [g]
    d_max_95 = 0.05 * weight * 1000
    sweat_rate = sw_tot

    def_dir = 0
    if theta != 0:
        # def_dir = 1 for unidirectional walking, def_dir = 0 for omni-directional walking
        def_dir = 1
    if walk_sp == 0:
        def_speed = 0
    else:
        def_speed = 1

    # radiating area dubois
    a_r_du = 0.7
    if posture == 2:
        a_r_du = 0.77
    if posture == 3:
        a_r_du = 0.67

    # evaluation of the max sweat rate as a function of the metabolic rate
    sw_max = (met - 32) * a_dubois
    if sw_max > 400:
        sw_max = 400
    if sw_max < 250:
        sw_max = 250
    if acclimatized >= 50:
        sw_max = sw_max * 1.25

    # max skin wettedness
    if acclimatized < 50:
        w_max = 0.85
    else:
        w_max = 1

    # static clothing insulation
    i_cl_st = clo * 0.155
    fcl = 1 + 0.3 * clo

    # Static boundary layer thermal insulation in quiet air
    i_a_st = 0.111

    # Total static insulation
    i_tot_st = i_cl_st + i_a_st / fcl
    if def_speed > 0:
        if def_dir == 1:  # Unidirectional walking
            v_r = abs(v - walk_sp * math.cos(3.14159 * theta / 180))
        else:  # Omni-directional walking IF
            if v < walk_sp:
                v_r = walk_sp
            else:
                v_r = v
    else:
        walk_sp = 0.0052 * (met - 58)
        if walk_sp > 0.7:
            walk_sp = 0.7
        v_r = v

    # Dynamic clothing insulation - correction for wind (Var) and walking speed
    v_ux = v_r
    if v_r > 3:
        v_ux = 3
    w_a_ux = walk_sp
    if walk_sp > 1.5:
        w_a_ux = 1.5
    # correction for the dynamic total dry thermal insulation at or above 0.6 clo
    corr_cl = 1.044 * math.exp(
        (0.066 * v_ux - 0.398) * v_ux + (0.094 * w_a_ux - 0.378) * w_a_ux
    )
    if corr_cl > 1:
        corr_cl = 1
    # correction for the dynamic total dry thermal insulation at 0 clo
    corr_ia = math.exp((0.047 * v_r - 0.472) * v_r + (0.117 * w_a_ux - 0.342) * w_a_ux)
    if corr_ia > 1:
        corr_ia = 1
    corr_tot = corr_cl
    if clo <= 0.6:
        corr_tot = ((0.6 - clo) * corr_ia + clo * corr_cl) / 0.6
    # total dynamic clothing insulation
    i_tot_dyn = i_tot_st * corr_tot
    # dynamic boundary layer thermal insulation
    i_a_dyn = corr_ia * i_a_st
    i_cl_dyn = i_tot_dyn - i_a_dyn / fcl
    # correction for the dynamic permeability index
    corr_e = (2.6 * corr_tot - 6.5) * corr_tot + 4.9
    im_dyn = i_mst * corr_e
    if im_dyn > 0.9:
        im_dyn = 0.9
    r_t_dyn = i_tot_dyn / im_dyn / 16.7
    t_exp = 28.56 + 0.115 * tdb + 0.641 * p_a  # expired air temperature
    # respiratory convective heat flow [W/m2]
    c_res = 0.001516 * met * (t_exp - tdb)
    # respiratory evaporative heat flow [W/m2]
    e_res = 0.00127 * met * (59.34 + 0.53 * tdb - 11.63 * p_a)
    z = 3.5 + 5.2 * v_r
    if v_r > 1:
        z = 8.7 * v_r**0.6

    # dynamic convective heat transfer coefficient
    hc_dyn = 2.38 * abs(t_sk - tdb) ** 0.25
    if z > hc_dyn:
        hc_dyn = z

    aux_r = 5.67e-08 * a_r_du
    f_cl_r = (1 - a_p) * 0.97 + a_p * f_r

    for time in range(1, duration + 1):
        t_sk0 = t_sk
        t_re0 = t_re
        t_cr0 = t_cr
        t_cr_eq0 = t_cr_eq
        t_sk_t_cr_wg0 = t_sk_t_cr_wg

        # equilibrium core temperature associated to the metabolic rate
        t_cr_eq_m = 0.0036 * met + 36.6
        # Core temperature at this minute, by exponential averaging
        t_cr_eq = t_cr_eq0 * const_t_eq + t_cr_eq_m * (1 - const_t_eq)
        # Heat storage associated with this core temperature increase during the last minute
        d_stored_eq = sp_heat * (t_cr_eq - t_cr_eq0) * (1 - t_sk_t_cr_wg0)
        # skin temperature prediction -- clothed model
        t_sk_eq_cl = 12.165 + 0.02017 * tdb + 0.04361 * tr + 0.19354 * p_a - 0.25315 * v
        t_sk_eq_cl = t_sk_eq_cl + 0.005346 * met + 0.51274 * t_re
        # nude model
        t_sk_eq_nu = 7.191 + 0.064 * tdb + 0.061 * tr + 0.198 * p_a - 0.348 * v
        t_sk_eq_nu = t_sk_eq_nu + 0.616 * t_re
        if clo >= 0.6:
            t_sk_eq = t_sk_eq_cl
        elif clo <= 0.2:
            t_sk_eq = t_sk_eq_nu
        else:
            t_sk_eq = t_sk_eq_nu + 2.5 * (t_sk_eq_cl - t_sk_eq_nu) * (clo - 0.2)

        # skin temperature [C]
        t_sk = t_sk0 * const_t_sk + t_sk_eq * (1 - const_t_sk)
        # Saturated water vapour pressure at the surface of the skin
        p_sk = 0.6105 * math.exp(17.27 * t_sk / (t_sk + 237.3))
        t_cl = tr + 0.1  # clothing surface temperature
        while True:
            # radiative heat transfer coefficient
            h_r = f_cl_r * aux_r * ((t_cl + 273) ** 4 - (tr + 273) ** 4) / (t_cl - tr)
            t_cl_new = (fcl * (hc_dyn * tdb + h_r * tr) + t_sk / i_cl_dyn) / (
                fcl * (hc_dyn + h_r) + 1 / i_cl_dyn
            )
            if abs(t_cl - t_cl_new) <= 0.001:
                break
            t_cl = (t_cl + t_cl_new) / 2

        convection = fcl * hc_dyn * (t_cl - tdb)
        radiation = fcl * h_r * (t_cl - tr)
        # maximum evaporative heat flow at the skin surface [W/m2]
        e_max = (p_sk - p_a) / r_t_dyn
        if e_max == 0:  # added this otherwise e_req / e_max cannot be calculated
            e_max = 0.001
        # required evaporative heat flow [W/m2]
        e_req = met - d_stored_eq - wme - c_res - e_res - convection - radiation
        # required skin wettedness
        w_req = e_req / e_max

        if e_req <= 0:
            e_req = 0
            sw_req = 0  # required sweat rate [W/m2]
        elif e_max <= 0:
            e_max = 0
            sw_req = sw_max
        elif w_req >= 1.7:
            sw_req = sw_max
        else:
            e_v_eff = 1 - w_req**2 / 2
            if w_req > 1:
                e_v_eff = (2 - w_req) ** 2 / 2
            sw_req = e_req / e_v_eff
            if sw_req > sw_max:
                sw_req = sw_max
        sweat_rate = sweat_rate * const_sw + sw_req * (1 - const_sw)

        if sweat_rate <= 0:
            e_p = 0  # predicted evaporative heat flow [W/m2]
            sweat_rate = 0
        else:
            k = e_max / sweat_rate
            wp = 1
            if k >= 0.5:
                wp = -k + math.sqrt(k * k + 2)
            if wp > w_max:
                wp = w_max
            e_p = wp * e_max

        # body heat storage rate [W/m2]
        d_storage = e_req - e_p + d_stored_eq
        t_cr_new = t_cr0
        while True:
            t_sk_t_cr_wg = 0.3 - 0.09 * (t_cr_new - 36.8)
            if t_sk_t_cr_wg > 0.3:
                t_sk_t_cr_wg = 0.3
            if t_sk_t_cr_wg < 0.1:
                t_sk_t_cr_wg = 0.1
            t_cr = (
                d_storage / sp_heat
                + t_sk0 * t_sk_t_cr_wg0 / 2
                - t_sk * t_sk_t_cr_wg / 2
            )
            t_cr = (t_cr + t_cr0 * (1 - t_sk_t_cr_wg0 / 2)) / (1 - t_sk_t_cr_wg / 2)
            if abs(t_cr - t_cr_new) <= 0.001:
                break
            t_cr_new = (t_cr_new + t_cr) / 2

        t_re = t_re0 + (2 * t_cr - 1.962 * t_re0 - 1.31) / 9
        if d_lim_t_re == 0 and t_re >= 38:
            d_lim_t_re = time
        sw_tot = sw_tot + sweat_rate + e_res
        sw_tot_g = sw_tot * 2.67 * a_dubois / 1.8 / 60
        if d_lim_loss_50 == 0 and sw_tot_g >= d_max_50:
            d_lim_loss_50 = time
        if d_lim_loss_95 == 0 and sw_tot_g >= d_max_95:
            d_lim_loss_95 = time
        if drink == 0:
            d_lim_loss_95 = d_lim_loss_95 * 0.6
            d_lim_loss_50 = d_lim_loss_95
    if d_lim_loss_50 == 0:
        d_lim_loss_50 = duration
    if d_lim_loss_95 == 0:
        d_lim_loss_95 = duration
    if d_lim_t_re == 0:
        d_lim_t_re = duration

    return (
        t_re,
        t_sk,
        t_cr,
        t_cr_eq,
        t_sk_t_cr_wg,
        sweat_rate,
        sw_tot_g,
        d_lim_loss_50,
        d_lim_loss_95,
        d_lim_t_re,
    )



import warnings
from typing import Union

from scipy import optimize

from pythermalcomfort.utilities import (
    units_converter,
)
from pythermalcomfort.models.set_tmp import set_tmp


def cooling_effect(
    tdb: Union[float, int],
    tr: Union[float, int],
    vr: Union[float, int],
    rh: Union[float, int],
    met: Union[float, int],
    clo: Union[float, int],
    wme=0,
    units="SI",
):
    """Returns the value of the Cooling Effect (`CE`_) calculated in compliance
    with the ASHRAE 55 2020 Standard [1]_. The `CE`_ of the elevated air speed
    is the value that, when subtracted equally from both the average air
    temperature and the mean radiant temperature, yields the same `SET`_ under
    still air as in the first `SET`_ calculation under elevated air speed. The
    cooling effect is calculated only for air speed higher than 0.1 m/s.

    .. _CE: https://en.wikipedia.org/wiki/Thermal_comfort#Cooling_Effect

    Parameters
    ----------
    tdb : float
        dry bulb air temperature, default in [°C] in [°F] if `units` = 'IP'
    tr : float
        mean radiant temperature, default in [°C] in [°F] if `units` = 'IP'
    vr : float
        relative air speed, default in [m/s] in [fps] if `units` = 'IP'

        Note: vr is the relative air speed caused by body movement and not the air
        speed measured by the air speed sensor. The relative air speed is the sum of the
        average air speed measured by the sensor plus the activity-generated air speed
        (Vag). Where Vag is the activity-generated air speed caused by motion of
        individual body parts. vr can be calculated using the function
        :py:meth:`pythermalcomfort.utilities.v_relative`.
    rh : float
        relative humidity, [%]
    met : float
        metabolic rate, [met]
    clo : float
        clothing insulation, [clo]

        Note: The activity as well as the air speed modify the insulation characteristics
        of the clothing and the adjacent air layer. Consequently the ISO 7730 states that
        the clothing insulation shall be corrected [2]_. The ASHRAE 55 Standard corrects
        for the effect of the body movement for met equal or higher than 1.2 met using
        the equation clo = Icl × (0.6 + 0.4/met) The dynamic clothing insulation, clo,
        can be calculated using the function
        :py:meth:`pythermalcomfort.utilities.clo_dynamic`.
    wme : float, default 0
        external work, [met]
    units : {'SI', 'IP'}     select the SI (International System of Units) or the IP (Imperial Units) system.

    Returns
    -------
    ce : float
        Cooling Effect, default in [°C] in [°F] if `units` = 'IP'

    Examples
    --------
    .. code-block:: python

        >>> from pythermalcomfort.models import cooling_effect
        >>> CE = cooling_effect(tdb=25, tr=25, vr=0.3, rh=50, met=1.2, clo=0.5)
        >>> print(CE)
        1.64

        >>> # for users who wants to use the IP system
        >>> CE = cooling_effect(tdb=77, tr=77, vr=1.64, rh=50, met=1, clo=0.6, units="IP")
        >>> print(CE)
        3.74

    Raises
    ------
    ValueError
        If the cooling effect could not be calculated
    """

    if units.lower() == "ip":
        tdb, tr, vr = units_converter(tdb=tdb, tr=tr, v=vr)

    if vr <= 0.1:
        return 0

    still_air_threshold = 0.1

    warnings.simplefilter("ignore")

    initial_set_tmp = set_tmp(
        tdb=tdb,
        tr=tr,
        v=vr,
        rh=rh,
        met=met,
        clo=clo,
        wme=wme,
        round=False,
        calculate_ce=True,
        limit_inputs=False,
    )

    def function(x):
        return (
            set_tmp(
                tdb - x,
                tr - x,
                v=still_air_threshold,
                rh=rh,
                met=met,
                clo=clo,
                wme=wme,
                round=False,
                calculate_ce=True,
                limit_inputs=False,
            )
            - initial_set_tmp
        )

    try:
        ce = optimize.brentq(function, 0.0, 40)
    except ValueError:
        ce = 0

    warnings.simplefilter("always")

    if ce == 0:
        warnings.warn(
            "Assuming cooling effect = 0 since it could not be calculated for this set"
            f" of inputs {tdb=}, {tr=}, {rh=}, {vr=}, {clo=}, {met=}",
            UserWarning,
        )

    if units.lower() == "ip":
        ce = ce / 1.8 * 3.28

    return round(ce, 2)
